@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2011, University of Amsterdam
+    Copyright (C): 1985-2013, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -129,24 +127,43 @@ visible(Ports) :-
 	map_bits(port_name, Ports, Old, New),
 	'$visible'(_, New).
 
-style_name(atom,	    2'0000001).
-style_name(singleton,	    2'0000010).
-style_name(discontiguous,   2'0001000).
-style_name(dynamic,	    2'0010000).
-style_name(charset,	    2'0100000).
+style_name(atom,	    0x0001).
+style_name(singleton,	    0x0042).		% semantic and syntactic
+style_name(discontiguous,   0x0008).
+style_name(dynamic,	    0x0010).
+style_name(charset,	    0x0020).
+style_name(no_effect,	    0x0080).
+style_name(var_branches,    0x0100).
 
+%%	style_check(+Spec) is nondet.
+
+style_check(Var) :-
+	var(Var), !,
+	'$instantiation_error'(Var).
 style_check(+string) :- !,
 	set_prolog_flag(double_quotes, string).
 style_check(-string) :- !,
 	set_prolog_flag(double_quotes, codes).
-style_check(?(string)) :- !,
-	current_prolog_flag(double_quotes, string).
+style_check(?(Style)) :- !,
+	(   var(Style)
+	->  enum_style_check(Style)
+	;   enum_style_check(Style)
+	->  true
+	).
 style_check(Spec) :-
 	'$style_check'(Old, Old),
 	map_bits(style_name, Spec, Old, New),
 	'$style_check'(_, New).
 
-%	prolog:debug_control_hook(+Action)
+enum_style_check(string) :-
+	current_prolog_flag(double_quotes, string).
+enum_style_check(Style) :-
+	'$style_check'(Bits, Bits),
+	style_name(Style, Bit),
+	Bit /\ Bits =\= 0.
+
+
+%%	prolog:debug_control_hook(+Action)
 %
 %	Allow user-hooks in the Prolog debugger interaction.  See the calls
 %	below for the provided hooks.  We use a single predicate with action
@@ -241,7 +258,16 @@ tag_list([H0|T0], F, [H1|T1]) :-
 %%	nospy(:Spec) is det.
 %%	nospyall is det.
 %
-%	Set/clear spy-points.
+%	Set/clear spy-points. A successfully set or cleared spy-point is
+%	reported using print_message/2, level  =informational=, with one
+%	of the following terms, where Spec is of the form M:Head.
+%
+%	    - spy(Spec)
+%	    - nospy(Spec)
+%
+%	@see	spy/1 and nospy/1 call the hook prolog:debug_control_hook/1
+%		to allow for alternative specifications of the thing to
+%		debug.
 
 spy(_:X) :-
 	var(X),
@@ -372,12 +398,16 @@ atom_prefix(Atom, Prefix) :-
 %	to create a file record without being interested in the time.
 
 source_file(File) :-
+	(   current_prolog_flag(access_level, user)
+	->  Level = user
+	;   true
+	),
 	(   ground(File)
-	->  (	'$time_source_file'(File, Time, user)
+	->  (	'$time_source_file'(File, Time, Level)
 	    ;	absolute_file_name(File, Abs),
-		'$time_source_file'(Abs, Time, user)
+		'$time_source_file'(Abs, Time, Level)
 	    ), !
-	;   '$time_source_file'(File, Time, user)
+	;   '$time_source_file'(File, Time, Level)
 	),
 	Time > 0.0.
 
@@ -474,6 +504,8 @@ prolog_load_context(script, Bool) :-
 	->  Bool = true
 	;   Bool = false
 	).
+prolog_load_context(variable_names, Bindings) :-
+	nb_current('$variable_names', Bindings).
 
 %%	unload_file(+File) is det.
 %
@@ -695,6 +727,8 @@ define_or_generate(Pred) :-
 	'$get_predicate_attribute'(Pred, noprofile, 1).
 '$predicate_property'(iso, Pred) :-
 	'$get_predicate_attribute'(Pred, iso, 1).
+'$predicate_property'(quasi_quotation_syntax, Pred) :-
+	'$get_predicate_attribute'(Pred, quasi_quotation_syntax, 1).
 
 system_undefined(user:prolog_trace_interception/4).
 system_undefined(user:prolog_exception_hook/4).
@@ -885,6 +919,22 @@ shell :-
 shell :-
 	shell('/bin/sh').
 
+%%	win_add_dll_directory(+AbsDir) is det.
+%
+%	Add AbsDir to the directories where  dependent DLLs are searched
+%	on Windows systems.
+
+:- if(current_prolog_flag(windows, true)).
+:- export(win_add_dll_directory/1).
+win_add_dll_directory(Dir) :-
+	win_add_dll_directory(Dir, _), !.
+win_add_dll_directory(Dir) :-
+	prolog_to_os_filename(Dir, OSDir),
+	getenv('PATH', Path0),
+	atomic_list_concat([Path0, OSDir], ';', Path),
+	setenv('PATH', Path).
+:- endif.
+
 		 /*******************************
 		 *	      SIGNALS		*
 		 *******************************/
@@ -902,11 +952,7 @@ on_signal(Signal, Old, New) :-
 	integer(Signal), !,
 	'$on_signal'(Signal, _Name, Old, New).
 on_signal(Signal, _Old, _New) :-
-	(   var(Signal)
-	->  Err = instantiation_error
-	;   Err = type_error(signal_name, Signal)
-	),
-	throw(error(Err, context(on_signal/3, _))).
+	'$type_error'(signal_name, Signal).
 
 %%	current_signal(?Name, ?SignalNumber, :Handler) is nondet.
 
