@@ -97,9 +97,10 @@ qsave_program(FileBase, Options0) :-
 	),
 	'$rc_open_archive'(File, RC),
 	make_header(RC, SaveClass, Options),
-	save_options(RC, [ init_file(InitFile)
-			 | Options
-			 ]),
+	save_options(RC, SaveClass,
+		     [ init_file(InitFile)
+		     | Options
+		     ]),
 	save_resources(RC, SaveClass),
 	'$rc_open'(RC, '$state', '$prolog', write, StateFd),
 	'$open_wic'(StateFd),
@@ -182,10 +183,11 @@ convert_option(Stack, Val, NewVal) :-	% stack-sizes are in K-bytes
 	->  NewVal = Val
 	;   NewVal is max(Min, Val*1024)
 	).
-convert_option(goal, Callable, Atom) :-
+convert_option(goal, Callable, Atom) :- !,
 	term_to_atom(Callable, Atom).
-convert_option(toplevel, Callable, Atom) :-
+convert_option(toplevel, Callable, Atom) :- !,
 	term_to_atom(Callable, Atom).
+convert_option(_, Value, Value).
 
 doption(Name) :- min_stack(Name, _).
 doption(goal).
@@ -195,7 +197,7 @@ doption(system_init_file).
 doption(class).
 doption(home).
 
-%%	save_options(+ArchiveHandle, +Options)
+%%	save_options(+ArchiveHandle, +SaveClass, +Options)
 %
 %	Save the options in the '$options'  resource. The home directory
 %	is saved for development saves,  so   it  keeps  refering to the
@@ -204,18 +206,15 @@ doption(home).
 %	The script-file (-s script) is not saved at all. I think this is
 %	fine to avoid a save-script loading itself.
 
-save_options(RC, Options) :-
+save_options(RC, SaveClass, Options) :-
 	'$rc_open'(RC, '$options', '$prolog', write, Fd),
 	(   doption(OptionName),
 	    '$option'(OptionName, OptionVal0),
-	        (   OptionName == home	% save home if not runtime
-		->  \+ memberchk(class(runtime), Options)
-		;   true
-		),
-	        OptTerm =.. [OptionName,OptionVal1],
+	        save_option_value(SaveClass, OptionName, OptionVal0, OptionVal1),
+	        OptTerm =.. [OptionName,OptionVal2],
 	        (   option(OptTerm, Options)
-		->  convert_option(OptionName, OptionVal1, OptionVal)
-		;   OptionVal = OptionVal0
+		->  convert_option(OptionName, OptionVal2, OptionVal)
+		;   OptionVal = OptionVal1
 		),
 	        format(Fd, '~w=~w~n', [OptionName, OptionVal]),
 	    fail
@@ -223,6 +222,13 @@ save_options(RC, Options) :-
 	),
 	close(Fd).
 
+%%	save_option_value(+SaveClass, +OptionName, +OptionValue, -FinalValue)
+
+save_option_value(Class,   class, _,     Class) :- !.
+save_option_value(runtime, home,  _,     _) :- !, fail.
+save_option_value(_,       _,     Value, Value).
+
+
 		 /*******************************
 		 *	     RESOURCES		*
 		 *******************************/
@@ -392,17 +398,18 @@ pred_attrib(Attrib, Head,
 	strip_module(Head, M, Term),
 	functor(Term, Name, Arity).
 
-attrib_name(dynamic,	   dynamic,	  1).
-attrib_name(volatile,	   volatile,	  1).
-attrib_name(thread_local,  thread_local,  1).
-attrib_name(multifile,	   multifile,	  1).
-attrib_name(public,	   public,	  1).
-attrib_name(transparent,   transparent,	  1).
-attrib_name(discontiguous, discontiguous, 1).
-attrib_name(notrace,	   trace,	  0).
-attrib_name(show_childs,   hide_childs,	  0).
-attrib_name(built_in,      system,	  1).
-attrib_name(nodebug,       hide_childs,	  1).
+attrib_name(dynamic,		    dynamic,		    1).
+attrib_name(volatile,		    volatile,		    1).
+attrib_name(thread_local,	    thread_local,	    1).
+attrib_name(multifile,		    multifile,		    1).
+attrib_name(public,		    public,		    1).
+attrib_name(transparent,	    transparent,	    1).
+attrib_name(discontiguous,	    discontiguous,	    1).
+attrib_name(notrace,		    trace,		    0).
+attrib_name(show_childs,	    hide_childs,	    0).
+attrib_name(built_in,		    system,		    1).
+attrib_name(nodebug,		    hide_childs,	    1).
+attrib_name(quasi_quotation_syntax, quasi_quotation_syntax, 1).
 
 
 save_attribute(P, Attribute) :-
@@ -519,10 +526,12 @@ save_prolog_flags :-
 save_prolog_flags.
 
 no_save_flag(argv).
+no_save_flag(os_argv).
 no_save_flag(access_level).
 no_save_flag(tty_control).
 no_save_flag(readline).
 no_save_flag(associated_file).
+no_save_flag(cpu_count).
 no_save_flag(hwnd).			% should be read-only, but comes
 					% from user-code
 
@@ -623,10 +632,8 @@ find_foreign_library(FileSpec, SharedObject) :-
 	    format(atom(Cmd), '"~w" -o "~w" "~w"',
 		   [ Strip, Stripped, File ]),
 	    shell(Cmd)
-	->  SharedObject = Stripped,
-	    Delete = true
-	;   SharedObject = File,
-	    Delete = false
+	->  SharedObject = Stripped
+	;   SharedObject = File
 	).
 
 
