@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        J.Wielemaker@uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2009, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -19,7 +20,7 @@
 
     You should have received a copy of the GNU General Public
     License along with this library; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
     As a special exception, if you link this library with other files,
     compiled with a Free Software compiler, to produce an executable, this
@@ -43,8 +44,33 @@
 :- use_module(library(error)).
 :- use_module(library(option)).
 
-:- initialization
-   load_foreign_library(foreign(json)).
+:- use_foreign_library(foreign(json)).
+
+:- predicate_options(json_read/3, 3,
+		     [ null(ground),
+		       true(ground),
+		       false(ground),
+		       value_string_as(oneof([atom,string]))
+		     ]).
+:- predicate_options(json_write/3, 3,
+		     [ indent(nonneg),
+		       step(positive_integer),
+		       tab(positive_integer),
+		       width(nonneg),
+		       null(ground),
+		       true(ground),
+		       false(ground)
+		     ]).
+:- predicate_options(is_json_term/2, 2,
+		     [ null(ground),
+		       true(ground),
+		       false(ground)
+		     ]).
+:- predicate_options(atom_json_term/3, 3,
+		     [ as(oneof([atom,string,codes])),
+		       pass_to(json_read/3, 3),
+		       pass_to(json_write/3, 3)
+		     ]).
 
 /** <module> Reading and writing JSON serialization
 
@@ -154,20 +180,27 @@ type_term(codes,  Result, codes(Result)).
 %		packed string object.  Please note that =codes= or
 %		=chars= would produce ambiguous output and is therefore
 %		not supported.
+%
+%	If json_read/3 encounters end-of-file before any real data it
+%	binds Term to the term @(end_of_file).
 
 json_read(Stream, Term) :-
 	default_json_options(Options),
 	json_value(Stream, Term, _, Options).
 
 json_read(Stream, Term, Options) :-
-	make_json_options(Options, OptionTerm),
+	make_json_options(Options, OptionTerm, _RestOptions),
 	json_value(Stream, Term, _, OptionTerm).
 
 
 json_value(Stream, Term, Next, Options) :-
 	get_code(Stream, C0),
 	ws(C0, Stream, C1),
-	json_term(C1, Stream, Term, Next, Options).
+	(   C1 == -1
+	->  Term = @(end_of_file),
+	    Next = -1
+	;   json_term(C1, Stream, Term, Next, Options)
+	).
 
 json_term(0'{, Stream, json(Pairs), Next, Options) :- !,
 	ws(Stream, C),
@@ -229,7 +262,7 @@ json_array(C0, Stream, [Value|Tail], Options) :-
 codes_to_type(atom, Codes, Atom) :-
 	atom_codes(Atom, Codes).
 codes_to_type(string, Codes, Atom) :-
-	string_to_list(Atom, Codes).
+	string_codes(Atom, Codes).
 codes_to_type(codes, Codes, Codes).
 
 json_string_as_atom(0'", Stream, Atom) :-
@@ -392,20 +425,20 @@ json_write(Stream, Term) :-
 	json_write(Stream, Term, []).
 json_write(Stream, Term, Options) :-
 	make_json_write_state(Options, State, Options1),
-	make_json_options(Options1, OptionTerm),
+	make_json_options(Options1, OptionTerm, _RestOptions),
 	json_write_term(Term, Stream, State, OptionTerm).
 
 json_write_term(Var, _, _, _) :-
 	var(Var), !,
 	instantiation_error(Var).
 json_write_term(json(Pairs), Stream, State, Options) :- !,
-	space_if_not_at_left_margin(Stream),
+	space_if_not_at_left_margin(Stream, State),
 	write(Stream, '{'),
 	(   json_write_state_width(State, Width),
 	    (   Width == 0
 	    ->  true
 	    ;   json_write_state_indent(State, Indent),
-		json_print_length(json(Pairs), Width, Indent, _)
+		json_print_length(json(Pairs), Options, Width, Indent, _)
 	    )
 	->  set_width_of_json_write_state(0, State, State2),
 	    write_pairs_hor(Pairs, Stream, State2, Options),
@@ -417,13 +450,13 @@ json_write_term(json(Pairs), Stream, State, Options) :- !,
 	).
 json_write_term(List, Stream, State, Options) :-
 	is_list(List), !,
-	space_if_not_at_left_margin(Stream),
+	space_if_not_at_left_margin(Stream, State),
 	write(Stream, '['),
 	(   json_write_state_width(State, Width),
 	    (   Width == 0
 	    ->  true
 	    ;   json_write_state_indent(State, Indent),
-		json_print_length(List, Width, Indent, _)
+		json_print_length(List, Options, Width, Indent, _)
 	    )
 	->  set_width_of_json_write_state(0, State, State2),
 	    write_array_hor(List, Stream, State2, Options),
@@ -490,7 +523,7 @@ write_array_hor([], _, _, _).
 write_array_hor([H|T], Stream, State, Options) :-
 	json_write_term(H, Stream, State, Options),
 	(   T == []
-	->  true
+	->  write(Stream, ' ')
 	;   write(Stream, ', '),
 	    write_array_hor(T, Stream, State, Options)
 	).
@@ -517,65 +550,80 @@ step_indent(State0, State) :-
 	NewIndent is Indent+Step,
 	set_indent_of_json_write_state(NewIndent, State0, State).
 
-space_if_not_at_left_margin(Stream) :-
-	line_position(Stream, 0), !.
-space_if_not_at_left_margin(Stream) :-
+space_if_not_at_left_margin(Stream, State) :-
+	line_position(Stream, LinePos),
+	(   LinePos == 0
+	;   json_write_state_indent(State, LinePos)
+	), !.
+space_if_not_at_left_margin(Stream, _) :-
 	put_char(Stream, ' ').
 
 
-%%	json_print_length(+Value, +Max, +Len0, +Len) is semidet.
+%%	json_print_length(+Value, +Options, +Max, +Len0, +Len) is semidet.
 %
 %	True if Len-Len0 is the print-length of Value on a single line
 %	and Len-Len0 =< Max.
 %
 %	@tbd	Escape sequences in strings are not considered.
 
-json_print_length(json(Pairs), Max, Len0, Len) :- !,
+json_print_length(json(Pairs), Options, Max, Len0, Len) :- !,
 	Len1 is Len0 + 2,
 	Len1 =< Max,
-	pairs_print_length(Pairs, Max, Len1, Len).
-json_print_length(Array, Max, Len0, Len) :-
+	pairs_print_length(Pairs, Options, Max, Len1, Len).
+json_print_length(Array, Options, Max, Len0, Len) :-
 	is_list(Array), !,
 	Len1 is Len0 + 2,
 	Len1 =< Max,
-	array_print_length(Array, Max, Len1, Len).
-json_print_length(Number, Max, Len0, Len) :-
+	array_print_length(Array, Options, Max, Len1, Len).
+json_print_length(Null, Options, Max, Len0, Len) :-
+  json_options_null(Options, Null), !,
+	Len is Len0 + 4,
+	Len =< Max.
+json_print_length(False, Options, Max, Len0, Len) :-
+  json_options_false(Options, False), !,
+	Len is Len0 + 5,
+	Len =< Max.
+json_print_length(True, Options, Max, Len0, Len) :-
+  json_options_true(Options, True), !,
+	Len is Len0 + 4,
+	Len =< Max.
+json_print_length(Number, _Options, Max, Len0, Len) :-
 	number(Number), !,
-	atom_length(Number, AL),
+	write_length(Number, AL, []),
 	Len is Len0 + AL,
 	Len =< Max.
-json_print_length(@(Id), Max, Len0, Len) :- !,
+json_print_length(@(Id), _Options, Max, Len0, Len) :- !,
 	atom_length(Id, IdLen),
 	Len is Len0+IdLen,
 	Len =< Max.
-json_print_length(String, Max, Len0, Len) :-
+json_print_length(String, _Options, Max, Len0, Len) :-
 	string_len(String, Len0, Len),
 	Len =< Max.
 
-pairs_print_length([], _, Len, Len).
-pairs_print_length([H|T], Max, Len0, Len) :-
-	pair_len(H, Max, Len0, Len1),
+pairs_print_length([], _, _, Len, Len).
+pairs_print_length([H|T], Options, Max, Len0, Len) :-
+	pair_len(H, Options, Max, Len0, Len1),
 	(   T == []
 	->  Len = Len1
 	;   Len2 is Len1 + 2,
 	    Len2 =< Max,
-	    pairs_print_length(T, Max, Len2, Len)
+	    pairs_print_length(T, Options, Max, Len2, Len)
 	).
 
-pair_len(Name=Value, Max, Len0, Len) :-
+pair_len(Name=Value, Options, Max, Len0, Len) :-
 	string_len(Name, Len0, Len1),
 	Len2 is Len1+2,
 	Len2 =< Max,
-	json_print_length(Value, Max, Len2, Len).
+	json_print_length(Value, Options, Max, Len2, Len).
 
-array_print_length([], _, Len, Len).
-array_print_length([H|T], Max, Len0, Len) :-
-	json_print_length(H, Max, Len0, Len1),
+array_print_length([], _, _, Len, Len).
+array_print_length([H|T], Options, Max, Len0, Len) :-
+	json_print_length(H, Options, Max, Len0, Len1),
 	(   T == []
 	->  Len = Len1
 	;   Len2 is Len1+2,
 	    Len2 =< Max,
-	    array_print_length(T, Max, Len2, Len)
+	    array_print_length(T, Options, Max, Len2, Len)
 	).
 
 string_len(String, Len0, Len) :-
@@ -599,7 +647,7 @@ is_json_term(Term) :-
 	is_json_term2(Options, Term).
 
 is_json_term(Term, Options) :-
-	make_json_options(Options, OptionTerm),
+	make_json_options(Options, OptionTerm, _RestOptions),
 	is_json_term2(OptionTerm, Term).
 
 is_json_term2(_, Var) :-
