@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2009-2012, University of Amsterdam
+    Copyright (C): 2009-2013, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -45,6 +45,7 @@
 	    object_edit_button//2,	% +Obj, +Options, //
 	    object_source_button//2,	% +Obj, +Options, //
 	    doc_resources//1,		% +Options
+	    ensure_doc_objects/1,	% +File
 					% Support other backends
 	    doc_file_objects/5,		% +FSpec, -File, -Objs, -FileOpts, +Opts
 	    existing_linked_file/2,	% +FileSpec, -Path
@@ -60,13 +61,15 @@
 	    file//2,			% +File, +Options, //
 	    include//3,			% +File, +Type, +Options //
 	    tags//1,			% +Tags, //
-	    term//2,			% +Term, +Bindings, //
+	    term//3,			% +Text, +Term, +Bindings, //
 	    file_header//2,		% +File, +Options, //
 	    objects//2,			% +Objects, +Options, //
 	    object_ref//2,		% +Object, +Options, //
 	    object_href/2,		% +Object, -URL
 	    object_page//2,		% +Object, +Options, //
-	    object_synopsis//2		% +Object, +Options, //
+	    object_page_header//2,	% +File, +Options, //
+	    object_synopsis//2,		% +Object, +Options, //
+	    object_page_footer//2	% +Object, +Options, //
 	  ]).
 :- use_module(library(lists)).
 :- use_module(library(option)).
@@ -100,7 +103,7 @@ extracting module doc_wiki.pl into HTML+CSS.
 */
 
 :- public
-	params//1,			% Called from \Term output created
+	args//1,			% Called from \Term output created
 	pred_dt//3,			% by the wiki renderer
 	section//2,
 	tag//2.
@@ -144,6 +147,7 @@ extracting module doc_wiki.pl into HTML+CSS.
 		       header(boolean),
 		       links(boolean),
 		       no_manual(boolean),
+		       try_manual(boolean),
 		       search_in(oneof([all,app,man])),
 		       search_match(oneof([name,summary])),
 		       search_options(boolean)
@@ -215,16 +219,27 @@ extracting module doc_wiki.pl into HTML+CSS.
 %		If =true=, provide edit buttons. Default, these buttons
 %		are suppressed.
 %
-%	@param File	Prolog file specification.
+%		* title(+Title)
+%		Specify the page title.  Default is the base name of the
+%		file.
+%
+%	@param File	Prolog file specification or xref source id.
 
 doc_for_file(FileSpec, Options) :-
-	absolute_file_name(FileSpec,
-			   [ file_type(prolog),
-			     access(read)
-			   ],
-			   File),
-	file_base_name(File, Base),
-	Title = Base,
+	(   xref_current_source(FileSpec)
+	->  true
+	;   absolute_file_name(FileSpec,
+			       [ file_type(prolog),
+				 access(read)
+			       ],
+			       File)
+	),
+	(   option(title(Title), Options)
+	->  true
+	;   atom(File)
+	->  file_base_name(File, Title)
+	;   format(atom(Title), '~w', [FileSpec]) % emergency
+	),
 	doc_write_page(
 	    pldoc(result),
 	    title(Title),
@@ -251,7 +266,7 @@ prolog_file(FileSpec, Options) -->
 	       \file_header(File, FileOptions)
 	     | \objects(Objects, FileOptions)
 	     ]),
-	undocumented(Objects, FileOptions).
+	undocumented(File, Objects, FileOptions).
 
 %%	doc_resources(+Options)// is det.
 %
@@ -284,38 +299,105 @@ doc_resources(Options) -->
 %
 %		* doc(PI:predicate_indicator, File:Line, Comment)
 %
+%	We distinguish three different states for FileSpec:
+%
+%	  1. File was cross-referenced with collection enabled.  All
+%	     information is in the xref database.
+%	  2. File was loaded. If comments are not loaded,
+%	     cross-reference the file, while _storing_ the comments
+%	     as the compiler would do.
+%	  3. Neither of the above.  In this case we cross-reference the
+%	     file.
+%
 %	@param FileSpec File specification as used for load_files/2.
 %	@param File	Prolog canonical filename
 
+doc_file_objects(FileSpec, File, Objects, FileOptions, Options) :-
+	xref_current_source(FileSpec),
+	xref_option(FileSpec, comments(collect)), !,
+	File = FileSpec,
+	findall(Object, xref_doc_object(File, Object), Objects0),
+	reply_file_objects(File, Objects0, Objects, FileOptions, Options).
 doc_file_objects(FileSpec, File, Objects, FileOptions, Options) :-
 	absolute_file_name(FileSpec, File,
 			   [ file_type(prolog),
 			     access(read)
 			   ]),
-	Pos = File:Line,
+	source_file(File), !,
 	ensure_doc_objects(File),
+	Pos = File:Line,
 	findall(Line-doc(Obj,Pos,Comment),
 		doc_comment(Obj, Pos, _, Comment), Pairs),
 	keysort(Pairs, ByLine),
 	pairs_values(ByLine, Objs0),
+	reply_file_objects(File, Objs0, Objects, FileOptions, Options).
+doc_file_objects(FileSpec, File, Objects, FileOptions, Options) :-
+	absolute_file_name(FileSpec, File,
+			   [ file_type(prolog),
+			     access(read)
+			   ]),
+	xref_source(File, [silent(true)]),
+	findall(Object, xref_doc_object(File, Object), Objects0),
+	reply_file_objects(File, Objects0, Objects, FileOptions, Options).
+
+
+reply_file_objects(File, Objs0, Objects, FileOptions, Options) :-
 	module_info(File, ModuleOptions, Options),
 	file_info(Objs0, Objs1, FileOptions, ModuleOptions),
-	doc_hide_private(Objs1, Objects, ModuleOptions).
+	doc_hide_private(Objs1, ObjectsSelf, ModuleOptions),
+	include_reexported(ObjectsSelf, Objects, File, FileOptions).
 
-%%	ensure_doc_objects(+File)
+include_reexported(SelfObjects, Objects, File, Options) :-
+	option(include_reexported(true), Options),
+	option(module(Module), Options),
+	option(public(Exports), Options),
+	select_undocumented(Exports, Module, SelfObjects, Undoc),
+	re_exported_doc(Undoc, File, Module, REObjs, _),
+	REObjs \== [], !,
+	append(SelfObjects, REObjs, Objects).
+include_reexported(Objects, Objects, _, _).
+
+
+%%	xref_doc_object(File, DocObject) is nondet.
+
+xref_doc_object(File, doc(M:module(Title),File:0,Comment)) :-
+	xref_comment(File, Title, Comment),
+	xref_module(File, M).
+xref_doc_object(File, doc(M:Name/Arity,File:0,Comment)) :-
+	xref_comment(File, Head, _Summary, Comment),
+	xref_module(File, Module),
+	strip_module(Module:Head, M, Plain),
+	functor(Plain, Name, Arity).
+
+%%	ensure_doc_objects(+File) is det.
 %
-%	Ensure we have documentation about  File.   If  the  file is not
-%	loaded, run the cross-referencer on it   to collect the comments
+%	Ensure we have documentation about File.  If we have no comments
+%	for the file because it was loaded before comment collection was
+%	enabled, run the cross-referencer on it  to collect the comments
 %	and meta-information.
 %
-%	@tbd	We could also run the cross-referencer if the file
-%		was loaded, but before we are collecting comments.
-%		Is that useful?
+%	@param File is a canonical filename that is loaded.
+
+:- dynamic
+	no_comments/2.
 
 ensure_doc_objects(File) :-
-	source_file(File), !.
+	source_file(File), !,
+	(   doc_file_has_comments(File)
+	->  true
+	;   no_comments(File, TimeChecked),
+	    time_file(File, TimeChecked)
+	->  true
+	;   xref_source(File, [silent(true), comments(store)]),
+	    retractall(no_comments(File, _)),
+	    (	doc_file_has_comments(File)
+	    ->	true
+	    ;	time_file(File, TimeChecked),
+		assertz(no_comments(File, TimeChecked))
+	    )
+	).
 ensure_doc_objects(File) :-
-	xref_source(File).
+	xref_source(File, [silent(true)]).
 
 %%	module_info(+File, -ModuleOptions, +OtherOptions) is det.
 %
@@ -398,8 +480,10 @@ private(Module:PI, _Options) :-
 multifile(Obj, _Options) :-
 	strip_module(user:Obj, Module, PI),
 	pi_to_head(PI, Head),
-	predicate_property(Module:Head, multifile).
-
+	(   predicate_property(Module:Head, multifile)
+	;   xref_module(Source, Module),
+	    xref_defined(Source, Head, multifile(_))
+	), !.
 
 pi_to_head(Var, _) :-
 	var(Var), !, fail.
@@ -428,7 +512,7 @@ file_header(File, Options) -->
 	},
 	file_title([Base, ' -- ', Title], File, Options),
 	{ is_structured_comment(Comment, Prefixes),
-	  string_to_list(Comment, Codes),
+	  string_codes(Comment, Codes),
 	  indented_lines(Codes, Prefixes, Lines),
 	  section_comment_header(Lines, _Header, Lines1),
 	  wiki_lines_to_dom(Lines1, [], DOM)
@@ -465,8 +549,10 @@ file_title(Title, File, Options) -->
 %	file is not loaded because we do  not want to load files through
 %	the documentation system.
 
-reload_button(File, _Base, _Options) -->
-	{ \+ source_file(File) }, !,
+reload_button(File, _Base, Options) -->
+	{ \+ source_file(File),
+	  \+ option(files(_), Options)
+	}, !,
 	html(span(class(file_anot), '[not loaded]')).
 reload_button(_File, Base, Options) -->
 	{ option(edit(true), Options), !,
@@ -577,6 +663,7 @@ object(Obj, Mode0, Mode, Options) -->
 		  doc_comment(Obj, Pos, _Summary, Comment),
 		  Pairs)
 	}, !,
+	{ b_setval(pldoc_object, Obj) },
 	object(Obj, Pairs, Mode0, Mode, Options).
 
 object(Obj, Pairs, Mode0, Mode, Options) -->
@@ -594,7 +681,7 @@ object(Obj, _Pairs, Mode, Mode, _Options) -->
 
 pred_dom(Obj, Options, Pos-Comment, DOM) :-
 	is_structured_comment(Comment, Prefixes),
-	string_to_list(Comment, Codes),
+	string_codes(Comment, Codes),
 	indented_lines(Codes, Prefixes, Lines),
 	strip_module(user:Obj, Module, _),
 	process_modes(Lines, Module, Pos, Modes, Args, Lines1),
@@ -648,26 +735,48 @@ pop_mode(Mode, [H|Rest0], Rest) -->
 	html_end(H),
 	pop_mode(Mode, Rest0, Rest).
 
-%%	undocumented(+Objects, +Options)// is det.
+%%	undocumented(+File, +Objects, +Options)// is det.
 %
 %	Describe undocumented predicates if the file is a module file.
 
-undocumented(Objs, Options) -->
+undocumented(File, Objs, Options) -->
 	{ memberchk(module(Module), Options),
 	  memberchk(public(Exports), Options),
 	  select_undocumented(Exports, Module, Objs, Undoc),
-	  Undoc \== []
+	  re_exported_doc(Undoc, File, Module, REObjs, ReallyUnDoc)
 	}, !,
-	html([ h2(class(undoc), 'Undocumented predicates'),
-	       p(['The following predicates are exported, but not ',
-		  'or incorrectly documented.'
-		 ]),
-	       dl(class(undoc),
-		  \undocumented_predicates(Undoc, Options))
-	     ]).
-undocumented(_, _) -->
+	re_exported_doc(REObjs, Options),
+	undocumented(ReallyUnDoc, Options).
+undocumented(_, _, _) -->
 	[].
 
+re_exported_doc([], _) --> !.
+re_exported_doc(Objs, Options) -->
+	reexport_header(Objs, Options),
+	objects(Objs, Options).
+
+reexport_header(_, Options) -->
+	{ option(reexport_header(true), Options, true)
+	}, !,
+	html([ h2(class(wiki), 'Re-exported predicates'),
+	       p([ 'The following predicates are re-exported from other ',
+		   'modules'
+		 ])
+	     ]).
+reexport_header(_, _) -->
+	[].
+
+undocumented([], _) --> !.
+undocumented(UnDoc, Options) -->
+	 html([ h2(class(undoc), 'Undocumented predicates'),
+		p(['The following predicates are exported, but not ',
+		   'or incorrectly documented.'
+		  ]),
+		dl(class(undoc),
+		   \undocumented_predicates(UnDoc, Options))
+	      ]).
+
+
 undocumented_predicates([], _) -->
 	[].
 undocumented_predicates([H|T], Options) -->
@@ -681,7 +790,7 @@ undocumented_pred(Name/Arity, Options) -->
 select_undocumented([], _, _, []).
 select_undocumented([PI|T0], M, Objs, [PI|T]) :-
 	is_pi(PI),
-	\+ in_doc(M:PI, Objs),
+	\+ in_doc(M:PI, Objs), !,
 	select_undocumented(T0, M, Objs, T).
 select_undocumented([_|T0], M, Objs, T) :-
 	select_undocumented(T0, M, Objs, T).
@@ -721,11 +830,29 @@ is_pi(_/_).
 is_pi(_//_).
 
 
+%%	re_exported_doc(+Undoc:list(pi), +File:atom, +Module:atom,
+%%			-ImportedDoc, -ReallyUnDoc:list(pi))
+
+re_exported_doc([], _, _, [], []).
+re_exported_doc([PI|T0], File, Module, [doc(Orig:PI,Pos,Comment)|ObjT], UnDoc) :-
+	pi_to_head(PI, Head),
+	(   predicate_property(Module:Head, imported_from(Orig))
+	->  true
+	;   xref_defined(File, Head, imported(File2)),
+	    ensure_doc_objects(File2),
+	    xref_module(File2, Orig)
+	),
+	doc_comment(Orig:PI, Pos, _, Comment), !,
+	re_exported_doc(T0, File, Module, ObjT, UnDoc).
+re_exported_doc([PI|T0], File, Module, REObj, [PI|UnDoc]) :-
+	re_exported_doc(T0, File, Module, REObj, UnDoc).
+
+
 		 /*******************************
 		 *	SINGLE OBJECT PAGE	*
 		 *******************************/
 
-%%	object_page(+Obj, +Options)// is det.
+%%	object_page(+Obj, +Options)// is semidet.
 %
 %	Generate an HTML page describing Obj.  The top presents the file
 %	the object is documented in and a search-form.  Options:
@@ -734,7 +861,8 @@ is_pi(_//_).
 %	    Show the navigation and search header.
 
 object_page(Obj, Options) -->
-	prolog:doc_object_page(Obj, Options).
+	prolog:doc_object_page(Obj, Options), !,
+	object_page_footer(Obj, Options).
 object_page(Obj, Options) -->
 	{ doc_comment(Obj, File:_Line, _Summary, _Comment)
 	}, !,
@@ -750,21 +878,35 @@ object_page(Obj, Options) -->
 		   \object_page_header(-, Options),
 		   \objects([Obj], [synopsis(true)|Options])
 		 ])
-	).
+	),
+	object_page_footer(Obj, Options).
 
 object_page_header(File, Options) -->
 	{ option(header(true), Options, true) }, !,
 	html(div(class(navhdr),
-		 [ \file_link(File),
+		 [ div(class(jump), \file_link(File)),
 		   div(class(search), \search_form(Options)),
 		   br(clear(right))
 		 ])).
 object_page_header(_, _) --> [].
 
-file_link(-) --> !.
+file_link(-) --> !,
+	places_menu(-).
 file_link(File) -->
-	html(div(class(jump),
-		 a(href(location_by_id(pldoc_doc)+File), File))).
+	{ file_directory_name(File, Dir)
+	},
+	places_menu(Dir),
+	html([ div(a(href(location_by_id(pldoc_doc)+File), File))
+	     ]).
+
+%%	object_page_footer(+Obj, +Options)// is det.
+%
+%	Call the hook prolog:doc_object_page_footer//2. This hook will
+%	be used to deal with annotations.
+
+object_page_footer(Obj, Options) -->
+	prolog:doc_object_page_footer(Obj, Options).
+object_page_footer(_, _) --> [].
 
 
 %%	object_synopsis(Obj, Options) is det.
@@ -931,24 +1073,24 @@ tag_title(tbd,    'To be done').
 tag_title(see,    'See also').
 tag_title(error,  'Errors').
 
-%%	params(+Params:list) is det.
+%%	args(+Params:list) is det.
 %
-%	Called from \params(List) created by   doc_wiki.pl.  Params is a
-%	list of param(Name, Descr).
+%	Called from \args(List) created by   doc_wiki.pl.  Params is a
+%	list of arg(Name, Descr).
 
-params(Params) -->
-	html([ dt(class=tag, 'Parameters:'),
-	       dd(table(class=paramlist,
-			\param_list(Params)))
+args(Params) -->
+	html([ dt(class=tag, 'Arguments:'),
+	       dd(table(class=arglist,
+			\arg_list(Params)))
 	     ]).
 
-param_list([]) -->
+arg_list([]) -->
 	[].
-param_list([H|T]) -->
-	param(H),
-	param_list(T).
+arg_list([H|T]) -->
+	argument(H),
+	arg_list(T).
 
-param(param(Name,Descr)) -->
+argument(arg(Name,Descr)) -->
 	html(tr([td(var(Name)), td(class=argdescr, ['- '|Descr])])).
 
 
@@ -957,7 +1099,7 @@ param(param(Name,Descr)) -->
 		 *******************************/
 
 section(Type, Title) -->
-	{ string_to_list(Title, Codes),
+	{ string_codes(Title, Codes),
 	  wiki_codes_to_dom(Codes, [], Content0),
 	  strip_leading_par(Content0, Content),
 	  make_section(Type, Content, HTML)
@@ -1286,21 +1428,21 @@ pred_det(Det) -->
 	html([' is ', b(class=det, Det)]).
 
 
-%%	term(+Term, +Bindings)// is det.
+%%	term(+Text, +Term, +Bindings)// is det.
 %
 %	Process the \term element as produced by doc_wiki.pl.
 %
 %	@tbd	Properly merge with pred_head//1
 
-term(Atom, []) -->
+term(_, Atom, []) -->
 	{ atomic(Atom) }, !,
 	html(span(class=functor, Atom)).
-term(Term, Bindings) -->
+term(_, Term, Bindings) -->
 	{ is_mode(Term is det),		% HACK. Bit too strict?
 	  bind_vars(Bindings)
 	}, !,
 	pred_head(Term).
-term(Term, Bindings) -->
+term(_, Term, Bindings) -->
 	{ bind_vars(Term, Bindings) },
 	argtype(Term).
 
@@ -1326,9 +1468,16 @@ predref(Term) -->
 
 predref(Obj, Options) -->
 	{ Obj = _:_,
-	  doc_comment(Obj, _, _, _)
+	  doc_comment(Obj, File:_Line, _, _),
+	  (   (   option(files(Map), Options)
+	      ->  memberchk(file(File,_), Map)
+	      ;   true
+	      )
+	  ->  object_href(Obj, HREF, Options)
+	  ;   manref(Obj, HREF, Options)
+	  )
 	}, !,
-	object_ref(Obj, [qualify(true)|Options]).
+	html(a(href(HREF), \object_link(Obj, [qualify(true)|Options]))).
 predref(M:Term, Options) --> !,
 	predref(Term, M, Options).
 predref(Term, Options) -->
@@ -1346,40 +1495,57 @@ predref(Name/Arity, _, Options) -->		% From packages
 	},
 	html(a([class=Category, href=HREF], [Name, /, Arity])).
 predref(Obj, Module, Options) -->		% Local
-	{ doc_comment(Module:Obj, _, _, _)
+	{ doc_comment(Module:Obj, File:_Line, _, _),
+	  (   option(files(Map), Options)
+	  ->  memberchk(file(File,_), Map)
+	  ;   true
+	  )
 	}, !,
 	object_ref(Module:Obj, Options).
-predref(Name/Arity, Module, _Options) -->
-	{ pred_href(Name/Arity, Module, HREF) }, !,
+predref(Name/Arity, Module, Options) -->
+	{ \+ option(files(_), Options),
+	  pred_href(Name/Arity, Module, HREF)
+	}, !,
 	html(a(href=HREF, [Name, /, Arity])).
-predref(Name//Arity, Module, _Options) -->
-	{ PredArity is Arity + 2,
+predref(Name//Arity, Module, Options) -->
+	{ \+ option(files(_), Options),
+	  PredArity is Arity + 2,
 	  pred_href(Name/PredArity, Module, HREF)
 	}, !,
 	html(a(href=HREF, [Name, //, Arity])).
-predref(Name/Arity, _, Options) -->		% From packages
-	{ prolog:doc_object_summary(Name/Arity, Category, _, _), !,
-	  manref(Name/Arity, HREF, Options)
+predref(PI, _, Options) -->		% From packages
+	{ canonical_pi(PI, CPI, HTML),
+	  (   option(files(_), Options)
+	  ->  Category = extmanual
+	  ;   prolog:doc_object_summary(CPI, Category, _, _)
+	  ),
+	  manref(CPI, HREF, Options)
 	},
-	html(a([class=Category, href=HREF], [Name, /, Arity])).
-predref(Name/Arity, _, _Options) --> !,
-	html(span(class=undef, [Name, /, Arity])).
-predref(Name//Arity, _, _Options) --> !,
-	html(span(class=undef, [Name, //, Arity])).
+	html(a([class=Category, href=HREF], HTML)).
+predref(PI, _, _Options) -->
+	{ canonical_pi(PI, _CPI, HTML)
+	}, !,
+	html(span(class=undef, HTML)).
 predref(Callable, Module, Options) -->
 	{ callable(Callable),
 	  functor(Callable, Name, Arity)
 	},
 	predref(Name/Arity, Module, Options).
 
+canonical_pi(Name/Arity, Name/Arity, [Name, /, Arity]) :-
+	atom(Name), integer(Arity), !.
+canonical_pi(Name//Arity, Name/Arity2, [Name, //, Arity]) :-
+	atom(Name), integer(Arity), !,
+	Arity2 is Arity+2.
+
 
 %%	manref(+NameArity, -HREF, +Options) is det.
 %
 %	Create reference to a manual page.  When generating files, this
 %	listens to the option man_server(+Server).
 
-manref(Name/Arity, HREF, Options) :-
-	format(string(FragmentId), '~w/~d', [Name, Arity]),
+manref(PI, HREF, Options) :-
+	predname(PI, PredName),
 	(   option(files(_Map), Options)
 	->  option(man_server(Server), Options,
 		   'http://www.swi-prolog.org/pldoc'),
@@ -1387,12 +1553,17 @@ manref(Name/Arity, HREF, Options) :-
 	    uri_data(path, Comp0, Path0),
 	    directory_file_path(Path0, man, Path),
 	    uri_data(path, Comp0, Path, Components),
-	    uri_query_components(Query, [predicate=FragmentId]),
+	    uri_query_components(Query, [predicate=PredName]),
 	    uri_data(search, Components, Query),
 	    uri_components(HREF, Components)
-	;   http_link_to_id(pldoc_man, [predicate=FragmentId], HREF)
+	;   http_link_to_id(pldoc_man, [predicate=PredName], HREF)
 	).
 
+predname(Name/Arity, PredName) :- !,
+	format(atom(PredName), '~w/~d', [Name, Arity]).
+predname(Module:Name/Arity, PredName) :- !,
+	format(atom(PredName), '~w:~w/~d', [Module, Name, Arity]).
+
 
 %%	pred_href(+NameArity, +Module, -HREF) is semidet.
 %
@@ -1483,7 +1654,10 @@ object_href(Obj, HREF) :-
 
 object_href(M:PI0, HREF, Options) :-
 	option(files(Map), Options),
-	module_property(M, file(File)),
+	(   module_property(M, file(File))
+	->  true
+	;   xref_module(File, M)
+	),
 	memberchk(file(File, DocFile), Map), !,
 	file_base_name(DocFile, LocalFile),	% TBD: proper directory index
 	expand_pi(PI0, PI),
@@ -1491,6 +1665,8 @@ object_href(M:PI0, HREF, Options) :-
 	uri_data(path, Components, LocalFile),
 	uri_data(fragment, Components, PIS),
 	uri_components(HREF, Components).
+object_href(Obj, HREF, _Options) :-
+	prolog:doc_object_href(Obj, HREF), !.
 object_href(Obj0, HREF, _Options) :-
 	localise_object(Obj0, Obj),
 	term_to_string(Obj, String),
@@ -1536,6 +1712,8 @@ term_to_string(Term, String) :-
 
 object_link(Obj, Options) -->
 	prolog:doc_object_link(Obj, Options), !.
+object_link(f(Name/Arity), _Options) --> !,
+	html([Name, /, Arity]).
 object_link(PI, Options) -->
 	{ is_pi(PI) }, !,
 	pi(PI, Options).
@@ -1744,7 +1922,13 @@ include(File, image, Options) -->
 			  type('image/svg+xml')
 			], Attrs)
 	},
-	html(object(Attrs, [])).
+	(   { option(caption(Caption), Options) }
+	->  html(div(class(figure),
+		     [ div(class(image), object(Attrs, [])),
+		       div(class(caption), Caption)
+		     ]))
+	;   html(object(Attrs, []))
+	).
 include(File, image, Options) -->
 	{ file_href(File, HREF, Options), !,
 	  include(image_attribute, Options, Attrs0),
@@ -1754,7 +1938,13 @@ include(File, image, Options) -->
 			  src(HREF)
 			], Attrs)
 	},
-	html(img(Attrs)).
+	(   { option(caption(Caption), Options) }
+	->  html(div(class(figure),
+		     [ div(class(image), img(Attrs)),
+		       div(class(caption), Caption)
+		     ]))
+	;   html(img(Attrs))
+	).
 include(File, wiki, _Options) -->	% [[file.txt]] is included
 	{ access_file(File, read), !,
 	  read_file_to_codes(File, String, []),
