@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2002-2012, University of Amsterdam
+    Copyright (C): 2002-2013, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -30,12 +30,13 @@
 
 :- module(rdf_persistency,
 	  [ rdf_attach_db/2,		% +Directory, +Options
-	    rdf_detach_db/0,		% +Detach current DB
+	    rdf_detach_db/0,		% +Detach current Graph
 	    rdf_current_db/1,		% -Directory
-	    rdf_persistency/2,		% +DB, +Bool
+	    rdf_persistency/2,		% +Graph, +Bool
 	    rdf_flush_journals/1,	% +Options
-	    rdf_journal_file/2,		% ?DB, ?JournalFile
-	    rdf_db_to_file/2		% ?DB, ?FileBase
+	    rdf_journal_file/2,		% ?Graph, ?JournalFile
+	    rdf_snapshot_file/2,	% ?Graph, ?SnapshotFile
+	    rdf_db_to_file/2		% ?Graph, ?FileBase
 	  ]).
 :- use_module(library(semweb/rdf_db)).
 :- use_module(library(filesex)).
@@ -44,7 +45,6 @@
 :- use_module(library(debug)).
 :- use_module(library(error)).
 :- use_module(library(thread)).
-:- use_module(library(pairs)).
 :- use_module(library(apply)).
 
 
@@ -97,13 +97,14 @@ delete the journals.
 	no_agc(0).
 
 :- predicate_options(rdf_attach_db/2, 2,
-		     [ concurrency(positive_integer),
+		     [ access(oneof([read_write,read_only])),
+		       concurrency(positive_integer),
 		       max_open_journals(positive_integer),
 		       silent(oneof([true,false,brief])),
 		       log_nested_transactions(boolean)
 		     ]).
 
-%%	rdf_attach_db(+Directory, +Options)
+%%	rdf_attach_db(+Directory, +Options) is det.
 %
 %	Start persistent operations using Directory   as  place to store
 %	files.   There are several cases:
@@ -116,6 +117,15 @@ delete the journals.
 %
 %	Options:
 %
+%		* access(+AccessMode)
+%		One of =auto= (default), =read_write= or
+%		=read_only=. Read-only access implies that the RDF
+%		store is not locked. It is read at startup and all
+%		modifications to the data are temporary. The default
+%		=auto= mode is =read_write= if the directory is
+%		writeable and the lock can be acquired.  Otherwise
+%		it reverts to =read_only=.
+%
 %		* concurrency(+Jobs)
 %		Number of threads to use for loading the initial
 %		database.  If not provided it is the number of CPUs
@@ -137,9 +147,59 @@ delete the journals.
 %		* log_nested_transactions(+Boolean)
 %		If =true=, nested _log_ transactions are added to the
 %		journal information.  By default (=false=), no log-term
-%		is added for nested transactions.
+%		is added for nested transactions.\\
+%
+%	@error existence_error(source_sink, Directory)
+%	@error permission_error(write, directory, Directory)
 
 rdf_attach_db(DirSpec, Options) :-
+	option(access(read_only), Options), !,
+	absolute_file_name(DirSpec,
+			   Directory,
+			   [ access(read),
+			     file_type(directory)
+			   ]),
+	rdf_detach_db,
+	assert(rdf_directory(Directory)),
+	assert_options(Options),
+	stop_monitor,		% make sure not to register load
+	no_agc(load_db).
+rdf_attach_db(DirSpec, Options) :-
+	option(access(read_write), Options), !,
+	rdf_attach_db_rw(DirSpec, Options).
+rdf_attach_db(DirSpec, Options) :-
+	absolute_file_name(DirSpec,
+			   Directory,
+			   [ access(exist),
+			     file_type(directory),
+			     file_errors(fail)
+			   ]), !,
+	(   access_file(Directory, write)
+	->  catch(rdf_attach_db_rw(Directory, Options), E, true),
+	    (	var(E)
+	    ->  true
+	    ;	E = error(permission_error(lock, rdf_db, _), _)
+	    ->	print_message(warning, E),
+		print_message(warning, rdf(read_only)),
+		rdf_attach_db(DirSpec, [access(read_only)|Options])
+	    ;	throw(E)
+	    )
+	;   print_message(warning,
+			  error(permission_error(write, directory, Directory))),
+	    print_message(warning, rdf(read_only)),
+	    rdf_attach_db_rw(Directory, Options)
+	).
+rdf_attach_db(DirSpec, Options) :-
+	catch(rdf_attach_db_rw(DirSpec, Options), E, true),
+	(   var(E)
+	->  true
+	;   print_message(warning, E),
+	    print_message(warning, rdf(read_only)),
+	    rdf_attach_db(DirSpec, [access(read_only)|Options])
+	).
+
+
+rdf_attach_db_rw(DirSpec, Options) :-
 	absolute_file_name(DirSpec,
 			   Directory,
 			   [ access(write),
@@ -158,7 +218,7 @@ rdf_attach_db(DirSpec, Options) :-
 	    at_halt(rdf_detach_db),
 	    start_monitor
 	).
-rdf_attach_db(DirSpec, Options) :-
+rdf_attach_db_rw(DirSpec, Options) :-
 	absolute_file_name(DirSpec,
 			   Directory,
 			   [ solutions(all)
@@ -168,6 +228,13 @@ rdf_attach_db(DirSpec, Options) :-
 	;   catch(make_directory(Directory), _, fail)
 	), !,
 	rdf_attach_db(Directory, Options).
+rdf_attach_db_rw(DirSpec, _) :-		% Generate an existence or
+	absolute_file_name(DirSpec,	% permission error
+			   Directory,
+			   [ access(exist),
+			     file_type(directory)
+			   ]),
+	permission_error(write, directory, Directory).
 
 
 assert_options([]).
@@ -184,6 +251,7 @@ option_type(max_open_journals(X),	must_be(positive_integer, X)).
 option_type(directory_levels(X),	must_be(positive_integer, X)).
 option_type(silent(X),	       must_be(oneof([true,false,brief]), X)).
 option_type(log_nested_transactions(X),	must_be(boolean, X)).
+option_type(access(X),	       must_be(oneof([read_write,read_only]), X)).
 
 
 %%	no_agc(:Goal)
@@ -266,22 +334,37 @@ rdf_flush_journal(DB, Options) :-
 
 load_db :-
 	rdf_directory(Dir),
-	get_time(Wall0),
-	statistics(cputime, T0),
-	load_prefixes(Dir),
-	find_dbs(Dir, DBs),
-	length(DBs, DBCount),
-	verbosity(DBCount, Silent),
-	make_goals(DBs, Silent, 1, DBCount, Goals),
 	concurrency(Jobs),
-	concurrent(Jobs, Goals, []),
-	statistics(cputime, T1),
+	cpu_stat_key(Jobs, StatKey),
+	get_time(Wall0),
+	statistics(StatKey, T0),
+	load_prefixes(Dir),
+	verbosity(Silent),
+	find_dbs(Dir, Graphs, SnapShots, Journals),
+	length(Graphs, GraphCount),
+	maplist(rdf_unload_graph, Graphs),
+	rdf_statistics(triples(Triples0)),
+	load_sources(snapshots, SnapShots, Silent, Jobs),
+	load_sources(journals, Journals, Silent, Jobs),
+	rdf_statistics(triples(Triples1)),
+	statistics(StatKey, T1),
 	get_time(Wall1),
 	T is T1 - T0,
 	Wall is Wall1 - Wall0,
+	Triples = Triples1 - Triples0,
 	message_level(Silent, Level),
-	print_message(Level, rdf(restore(attached(DBCount, T/Wall)))).
+	print_message(Level, rdf(restore(attached(GraphCount, Triples, T/Wall)))).
 
+load_sources(_, [], _, _) :- !.
+load_sources(Type, Sources, Silent, Jobs) :-
+	length(Sources, Count),
+	RunJobs is min(Count, Jobs),
+	print_message(informational, rdf(restoring(Type, Count, RunJobs))),
+	make_goals(Sources, Silent, 1, Count, Goals),
+	concurrent(RunJobs, Goals, []).
+
+
+%%	make_goals(+DBs, +Silent, +Index, +Total, -Goals)
 
 make_goals([], _, _, _, []).
 make_goals([DB|T0], Silent, I,  Total,
@@ -289,9 +372,12 @@ make_goals([DB|T0], Silent, I,  Total,
 	I2 is I + 1,
 	make_goals(T0, Silent, I2, Total, T).
 
-verbosity(_DBCount, Silent) :-
+verbosity(Silent) :-
 	rdf_option(silent(Silent)), !.
-verbosity(_DBCount, brief).
+verbosity(Silent) :-
+	current_prolog_flag(verbose, silent), !,
+	Silent = true.
+verbosity(brief).
 
 
 %%	concurrency(-Jobs)
@@ -305,48 +391,57 @@ concurrency(Jobs) :-
 	Jobs > 0, !.
 concurrency(1).
 
+cpu_stat_key(1, cputime) :- !.
+cpu_stat_key(_, process_cputime).
 
-%%	find_dbs(+Dir, -DBs:list(atom), -Depth) is det.
+
+%%	find_dbs(+Dir, -Graphs, -SnapBySize, -JournalBySize) is det.
 %
-%	DBs is a  list  of  database   (named  graph)  names,  sorted in
-%	increasing file-size. Small files  are   loaded  first  as these
-%	typically contain the schemas and we   want  to avoid re-hashing
-%	large databases due to added rdfs:subPropertyOf triples.
+%	Scan the persistent database and return a list of snapshots and
+%	journals, both sorted by file-size.  Each term is of the form
 %
-%	@param Depth is the depth at which all files have been found.
+%	  ==
+%	  db(Size, Ext, DB, DBFile, Depth)
+%	  ==
 
-find_dbs(Dir, DBs) :-
-	find_dbs(Dir, DBs, Depth),
-	(   var(Depth)			% empty DB
-	->  true
-	;   length(DBs, Count),
-	    LevelsRequired is floor(log(Count)/log(256)),
-	    Depth < LevelsRequired
-	->  print_message(informational, rdf(reindex(Count, LevelsRequired))),
-	    reindex_db(Dir, LevelsRequired),
-	    retractall(rdf_option(directory_levels(_))),
-	    assertz(rdf_option(directory_levels(LevelsRequired)))
-	;   retractall(rdf_option(directory_levels(_))),
-	    assertz(rdf_option(directory_levels(Depth)))
-	).
-
-find_dbs(Dir, DBs, Depth) :-
+find_dbs(Dir, Graphs, SnapBySize, JournalBySize) :-
 	directory_files(Dir, Files),
 	phrase(scan_db_files(Files, Dir, '.', 0), Scanned),
-	sort(Scanned, ByDB),
-	join_snapshot_and_journals(ByDB, BySize),
-	keysort(BySize, SortedBySize),
-	pairs_values(SortedBySize, DBs),
-	(   maplist(depth(Depth), Scanned)
-	->  true
-	;   length(DBs, Count),
-	    Levels is floor(log(Count)/log(256)),
-	    print_message(warning, rdf(fix_index(Levels))),
-	    reindex_db(Dir, Levels)
+	maplist(db_graph, Scanned, UnsortedGraphs),
+	sort(UnsortedGraphs, Graphs),
+	(   consider_reindex_db(Dir, Graphs, Scanned)
+	->  find_dbs(Dir, Graphs, SnapBySize, JournalBySize)
+	;   partition(db_is_snapshot, Scanned, Snapshots, Journals),
+	    sort(Snapshots, SnapBySize),
+	    sort(Journals, JournalBySize)
 	).
 
-depth(Depth, DB) :-
-	arg(4, DB, Depth).
+consider_reindex_db(Dir, Graphs, Scanned) :-
+	length(Graphs, Count),
+	Count > 0,
+	DepthNeeded is floor(log(Count)/log(256)),
+	(   maplist(depth_db(DepthNow), Scanned)
+	->  (   DepthNeeded > DepthNow
+	    ->	true
+	    ;	retractall(rdf_option(directory_levels(_))),
+		assertz(rdf_option(directory_levels(DepthNow))),
+		fail
+	    )
+	;   true
+	),
+	reindex_db(Dir, DepthNeeded).
+
+db_is_snapshot(Term) :-
+	arg(2, Term, trp).
+
+db_graph(Term, DB) :-
+	arg(3, Term, DB).
+
+db_file_name(Term, File) :-
+	arg(4, Term, File).
+
+depth_db(Depth, DB) :-
+	arg(5, DB, Depth).
 
 %%	scan_db_files(+Files, +Dir, +Prefix, +Depth)// is det.
 %
@@ -366,7 +461,7 @@ scan_db_files([File|T], Dir, Prefix, Depth) -->
 	  directory_file_path(Dir, DBFile, AbsFile),
 	  size_file(AbsFile, Size)
 	},
-	[ db(DB, Size, DBFile, Depth) ],
+	[ db(Size, Ext, DB, AbsFile, Depth) ],
 	scan_db_files(T, Dir, Prefix, Depth).
 scan_db_files([D|T], Dir, Prefix, Depth) -->
 	{ directory_file_path(Prefix, D, SubD),
@@ -387,32 +482,47 @@ nofollow(..).
 db_extension(trp).
 db_extension(jrn).
 
-join_snapshot_and_journals([], []).
-join_snapshot_and_journals([db(DB,S0,_,_)|T0], [S-DB|T]) :- !,
-	same_db(DB, T0, T1, S0, S),
-	join_snapshot_and_journals(T1, T).
-
-same_db(DB, [db(DB,S1,_,_)|T0], T, S0, S) :- !,
-	S2 is S0+S1,
-	same_db(DB, T0, T, S2, S).
-same_db(_, L, L, S, S).
-
-
 :- public load_source/4.		% called through make_goals/5
 
-%%	load_source(+DB, +Silent, +Nth, +Total) is det.
+load_source(DB, Silent, Nth, Total) :-
+	db_file_name(DB, File),
+	db_graph(DB, Graph),
+	message_level(Silent, Level),
+	graph_triple_count(Graph, Count0),
+	statistics(cputime, T0),
+	(   db_is_snapshot(DB)
+	->  print_message(Level, rdf(restore(Silent, snapshot(Graph, File)))),
+	    rdf_load_db(File)
+	;   print_message(Level, rdf(restore(Silent, journal(Graph, File)))),
+	    load_journal(File, Graph)
+	),
+	statistics(cputime, T1),
+	T is T1 - T0,
+	graph_triple_count(Graph, Count1),
+	Count is Count1 - Count0,
+	print_message(Level, rdf(restore(Silent,
+					 done(Graph, T, Count, Nth, Total)))).
+
+
+graph_triple_count(Graph, Count) :-
+	rdf_statistics(triples_by_graph(Graph, Count)), !.
+graph_triple_count(_, 0).
+
+
+%%	attach_graph(+Graph, +Options) is det.
 %
 %	Load triples and reload  journal   from  the  indicated snapshot
 %	file.
-%
-%	@param Silent One of =false=, =true= or =brief=
 
-load_source(DB, Silent, Nth, Total) :-
-	message_level(Silent, Level),
-	db_files(DB, SnapshotFile, JournalFile),
-	rdf_retractall(_,_,_,DB),
+attach_graph(Graph, Options) :-
+	(   option(silent(true), Options)
+	->  Level = silent
+	;   Level = informational
+	),
+	db_files(Graph, SnapshotFile, JournalFile),
+	rdf_retractall(_,_,_,Graph),
 	statistics(cputime, T0),
-	print_message(Level, rdf(restore(Silent, source(DB, Nth, Total)))),
+	print_message(Level, rdf(restore(Silent, Graph))),
 	db_file(SnapshotFile, AbsSnapShot),
 	(   exists_file(AbsSnapShot)
 	->  print_message(Level, rdf(restore(Silent, snapshot(SnapshotFile)))),
@@ -421,17 +531,17 @@ load_source(DB, Silent, Nth, Total) :-
 	),
 	(   exists_db(JournalFile)
 	->  print_message(Level, rdf(restore(Silent, journal(JournalFile)))),
-	    load_journal(JournalFile, DB)
+	    load_journal(JournalFile, Graph)
 	;   true
 	),
 	statistics(cputime, T1),
 	T is T1 - T0,
-	(   rdf_statistics(triples_by_file(DB, Count))
+	(   rdf_statistics(triples_by_graph(Graph, Count))
 	->  true
 	;   Count = 0
 	),
 	print_message(Level, rdf(restore(Silent,
-					 done(DB, T, Count, Nth, Total)))).
+					 done(Graph, T, Count)))).
 
 message_level(true, silent) :- !.
 message_level(_, informational).
@@ -447,11 +557,13 @@ message_level(_, informational).
 %	named graph.
 
 load_journal(File, DB) :-
-	open_db(File, read, In, []),
-	call_cleanup((  read(In, T0),
-			process_journal(T0, In, DB)
-		     ),
-		     close(In)).
+	rdf_create_graph(DB),
+	setup_call_cleanup(
+	    open(File, read, In, [encoding(utf8)]),
+	    ( read(In, T0),
+	      process_journal(T0, In, DB)
+	    ),
+	    close(In)).
 
 process_journal(end_of_file, _, _) :- !.
 process_journal(Term, In, DB) :-
@@ -513,6 +625,19 @@ rdf_persistency(DB, true) :-
 	;   true
 	).
 
+%%	rdf_db:property_of_graph(?Property, +Graph) is nondet.
+%
+%	Extend rdf_graph_property/2 with new properties.
+
+:- multifile
+	rdf_db:property_of_graph/2.
+
+rdf_db:property_of_graph(persistent(State), Graph) :-
+	(   blocked_db(Graph, persistency)
+	->  State = false
+	;   State = true
+	).
+
 
 %%	start_monitor is det.
 %%	stop_monitor is det.
@@ -582,11 +707,16 @@ monitor(update(S,P,O,DB,Action)) :-
 	    format(Fd, '~q.~n', [update(S,P,O,Action)]),
 	    sync_journal(DB, Fd)
 	).
-monitor(load(BE, Id)) :-
-	(   BE == begin
-	->  push_state(Id)
-	;   sync_state(Id)
+monitor(load(BE, _DumpFileURI)) :-
+	(   BE = end(Graphs)
+	->  sync_loaded_graphs(Graphs)
+	;   true
 	).
+monitor(create_graph(Graph)) :-
+	\+ blocked_db(Graph, _),
+	journal_fd(Graph, Fd),
+	open_transaction(Graph, Fd),
+	sync_journal(Graph, Fd).
 monitor(transaction(BE, Id)) :-
 	monitor_transaction(Id, BE).
 
@@ -746,38 +876,13 @@ end_transactions([DB:Id|T], DBs, N) :-
 	end_transactions(T, DBs, N).
 
 
-%	State  handling.  We  use   this    for   trapping   changes  by
-%	rdf_load_db/1. In theory, loading such files  can add triples to
-%	multiple sources. In practice this rarely   happens. We save the
-%	current state and sync all  files   that  have changed. The only
-%	drawback of this approach is that loaded files spreading triples
-%	over multiple databases cause all these   databases  to be fully
-%	synchronised. This shouldn't happen very often.
+%%	sync_loaded_graphs(+Graphs)
+%
+%	Called after a binary triple has been loaded that added triples
+%	to the given graphs.
 
-:- dynamic
-	pre_load_state/2.
-
-push_state(Id) :-
-	get_state(State),
-	asserta(pre_load_state(Id, State)).
-
-get_state(State) :-
-	findall(DB-MD5, (rdf_graph(DB), rdf_md5(DB, MD5)), State0),
-	keysort(State0, State).
-
-sync_state(Id) :-
-	retract(pre_load_state(Id, PreState)),
-	get_state(AfterState),
-	sync_state(AfterState, PreState).
-
-sync_state([], _).
-sync_state([DB-MD5|TA], Pre) :-
-	(   memberchk(DB-MD5P, Pre),
-	    MD5P == MD5
-	->  true
-	;   create_db(DB)
-	),
-	sync_state(TA, Pre).
+sync_loaded_graphs(Graphs) :-
+	maplist(create_db, Graphs).
 
 
 		 /*******************************
@@ -863,7 +968,7 @@ close_journal_(DB) :-
 	;   true
 	).
 
-%	close_journals
+%%	close_journals
 %
 %	Close all open journals.
 
@@ -929,8 +1034,8 @@ delete_db_(DB) :-
 lock_db(Dir) :-
 	lockfile(Dir, File),
 	catch(open(File, update, Out, [lock(write), wait(false)]),
-	      error(permission_error(lock, _, _), _),
-	      locked_error(Dir)),
+	      error(permission_error(Access, _, _), _),
+	      locked_error(Access, Dir)),
 	(   current_prolog_flag(pid, PID)
 	->  true
 	;   PID = 0			% TBD: Fix in Prolog
@@ -946,9 +1051,9 @@ lock_db(Dir) :-
 	flush_output(Out),
 	set_end_of_stream(Out),
 	assert(rdf_lock(Dir, lock(Out, File))),
-	at_halt(unlock_db(Out, File)).
+	at_halt(unlock_db(Dir)).
 
-locked_error(Dir) :-
+locked_error(lock, Dir) :-
 	lockfile(Dir, File),
 	(   catch(read_file_to_terms(File, Terms, []), _, fail),
 	    Terms = [locked(Args)]
@@ -956,6 +1061,9 @@ locked_error(Dir) :-
 	;   Context = context(_, 'Database is in use')
 	),
 	throw(error(permission_error(lock, rdf_db, Dir), Context)).
+locked_error(open, Dir) :-
+	throw(error(permission_error(lock, rdf_db, Dir),
+		    context(_, 'Lock file cannot be opened'))).
 
 %%	unlock_db(+Dir) is det.
 %%	unlock_db(+Stream, +File) is det.
@@ -963,6 +1071,7 @@ locked_error(Dir) :-
 unlock_db(Dir) :-
 	retract(rdf_lock(Dir, lock(Out, File))), !,
 	unlock_db(Out, File).
+unlock_db(_).
 
 unlock_db(Out, File) :-
 	close(Out),
@@ -1051,24 +1160,34 @@ db_abs_files(DB, Snapshot, Journal) :-
 	db_file(Journal0, Journal).
 
 
-%%	rdf_journal_file(+DB, -File) is semidet.
-%%	rdf_journal_file(-DB, -File) is nondet.
+%%	rdf_journal_file(+Graph, -File) is semidet.
+%%	rdf_journal_file(-Graph, -File) is nondet.
 %
-%	True if File is the absolute  file   name  of  an existing named
-%	graph DB.
-%
-%	@tbd	Avoid using private rdf_db:rdf_graphs_/1.
+%	True if File the name of the existing journal file for Graph.
 
-rdf_journal_file(DB, Journal) :-
-	(   var(DB)
-	->  rdf_db:rdf_graphs_(All),	% also pick the empty graphs
-	    member(DB, All)
+rdf_journal_file(Graph, Journal) :-
+	(   var(Graph)
+	->  rdf_graph(Graph)
 	;   true
 	),
-	db_abs_files(DB, _Snapshot, Journal),
+	db_abs_files(Graph, _Snapshot, Journal),
 	exists_file(Journal).
 
 
+%%	rdf_snapshot_file(+Graph, -File) is semidet.
+%%	rdf_snapshot_file(-Graph, -File) is nondet.
+%
+%	True if File the name of the existing snapshot file for Graph.
+
+rdf_snapshot_file(Graph, Snapshot) :-
+	(   var(Graph)
+	->  rdf_graph(Graph)	% also pick the empty graphs
+	;   true
+	),
+	db_abs_files(Graph, Snapshot, _Journal),
+	exists_file(Snapshot).
+
+
 %%	rdf_db_to_file(+DB, -File) is det.
 %%	rdf_db_to_file(-DB, +File) is det.
 %
@@ -1116,7 +1235,8 @@ url_encode(Enc) -->
 	(   "\r\n"
 	;   "\n"
 	), !,
-	{ append("%0D%0A", T, Enc)
+	{ string_codes("%0D%0A", Codes),
+	  append(Codes, T, Enc)
 	},
 	url_encode(T).
 url_encode([]) -->
@@ -1145,14 +1265,10 @@ no_enc_extra(0'_) --> "_".
 		 *	       REINDEX		*
 		 *******************************/
 
-%%	reindex_db(+Dir)
+%%	reindex_db(+Dir, +Levels)
 %
 %	Reindex the database by creating intermediate directories.
 
-reindex_db(Dir) :-
-	directory_levels(Levels),
-	reindex_db(Dir, Levels).
-
 reindex_db(Dir, Levels) :-
 	directory_files(Dir, Files),
 	reindex_files(Files, Dir, '.', 0, Levels),
@@ -1286,41 +1402,55 @@ time_stamp(Int) :-
 prolog:message(rdf(Term)) -->
 	message(Term).
 
-message(restore(attached(Graphs, Time/Wall))) -->
+message(restoring(Type, Count, Jobs)) -->
+	[ 'Restoring ~D ~w using ~D concurrent workers'-[Count, Type, Jobs] ].
+message(restore(attached(Graphs, Triples, Time/Wall))) -->
 	{ catch(Percent is round(100*Time/Wall), _, Percent = 0) },
-	[ 'Attached ~D graphs in ~2f seconds (~d% CPU = ~2f sec.)'-
-	  [Graphs, Wall, Percent, Time] ].
+	[ 'Loaded ~D graphs (~D triples) in ~2f sec. (~d% CPU = ~2f sec.)'-
+	  [Graphs, Triples, Wall, Percent, Time] ].
+% attach_graph/2
 message(restore(true, Action)) --> !,
 	silent_message(Action).
 message(restore(brief, Action)) --> !,
 	brief_message(Action).
-message(restore(_, source(DB, Nth, Total))) -->
-	{ file_base_name(DB, Base) },
-	[ 'Restoring ~w ... (~D of ~D graphs) '-[Base, Nth, Total], flush ].
+message(restore(_, Graph)) -->
+	[ 'Restoring ~p ... '-[Graph], flush ].
 message(restore(_, snapshot(_))) -->
 	[ at_same_line, '(snapshot) '-[], flush ].
 message(restore(_, journal(_))) -->
 	[ at_same_line, '(journal) '-[], flush ].
-message(restore(_, done(_, Time, Count, _Nth, _Total))) -->
+message(restore(_, done(_, Time, Count))) -->
 	[ at_same_line, '~D triples in ~2f sec.'-[Count, Time] ].
+% load_source/4
+message(restore(_, snapshot(G, _))) -->
+	[ 'Restoring ~p	(snapshot)'-[G], flush ].
+message(restore(_, journal(G, _))) -->
+	[ 'Restoring ~p	(journal)'-[G], flush ].
+message(restore(_, done(_, Time, Count))) -->
+	[ at_same_line, '~D triples in ~2f sec.'-[Count, Time] ].
+% journal handling
 message(update_failed(S,P,O,Action)) -->
 	[ 'Failed to update <~p ~p ~p> with ~p'-[S,P,O,Action] ].
+% directory reindexing
 message(reindex(Count, Depth)) -->
 	[ 'Restructuring database with ~d levels (~D graphs)'-[Depth, Count] ].
 message(reindex(Depth)) -->
 	[ 'Fixing database directory structure (~d levels)'-[Depth] ].
+message(read_only) -->
+	[ 'Cannot write persistent store; continuing in read-only mode.', nl,
+	  'All changes to the RDF store will be lost if this process terminates.'
+	].
 
 silent_message(_Action) --> [].
 
-brief_message(source(DB, Nth, Total)) -->
-	{ file_base_name(DB, Base) },
+brief_message(done(Graph, _Time, _Count, Nth, Total)) -->
+	{ file_base_name(Graph, Base) },
 	[ at_same_line,
-	  '\r~w~`.t ~D of ~D graphs~72|'-[Base, Nth, Total],
+	  '\r~p~`.t ~D of ~D graphs~72|'-[Base, Nth, Total],
 	  flush
 	].
-brief_message(snapshot(_File)) --> [].
-brief_message(journal(_File))  --> [].
-brief_message(done(_DB, _Time, _Count, _Nth, _Total)) --> [].
+brief_message(_) --> [].
+
 
 prolog:message_context(rdf_locked(Args)) -->
 	{ memberchk(time(Time), Args),
