@@ -48,7 +48,7 @@
 	   , forall/2
 	   , reverse/2
 	   , source_location/2
-	   , string_to_list/2
+	   , string_codes/2
 	   , append/3
 	   , atom_concat/3
 	   , between/3
@@ -64,9 +64,6 @@
 	verbose/0,
 	recording/2.			% items recorded
 
-:- if(current_predicate(index/1)).
-:- index(attribute(1,1,0)).
-:- endif.
 :- if(exists_source(library(quintus))).
 :- use_module(library(quintus), [genarg/3]).
 :- endif.
@@ -146,12 +143,15 @@ do_term_expand((Head :- Body), _) :-	% check for :- instead of :-> or :<-
 	fail.
 
 
-is_string(0) :- !, fail.		% catch variables
-is_string([]).
-is_string([H|T]) :-
-	integer(H),
-	between(0, 255, H),
-	is_string(T).
+%%	is_string(@Doc) is semidet.
+%
+%	See whether Doc may have originated from "..."
+
+is_string(Doc) :-
+	string(Doc), !.
+is_string(Doc) :-
+	is_list(Doc),
+	catch(string_codes(Doc, _), _, fail).
 
 typed_head(T) :-
 	functor(T, _, Arity),
@@ -300,6 +300,8 @@ do_expand((:- pce_end_class),
 	pop_class.
 do_expand((:- pce_end_class), []) :-
 	pce_error(no_class_to_end).
+do_expand((:- use_class_template(_)), []) :-
+	current_prolog_flag(xref, true), !.
 do_expand((:- use_class_template(Template)), []) :-
 	used_class_template(Template), !.
 do_expand((:- use_class_template(Template)),
@@ -752,7 +754,7 @@ pce_ifhostproperty(string, [
 (pce_summary(String, String) :-
 	string(String), !),
 (pce_summary(List, String) :-
-	string_to_list(String, List))]).
+	string_codes(String, List))]).
 pce_summary(List, string(List)).
 
 
@@ -813,7 +815,7 @@ to_atom(Atom, Atom) :-
 	atom(Atom), !.
 to_atom(Term, Atom) :-
 	ground(Term), !,
-	phrase(pce_type_description(Term), Chars, []),
+	phrase(pce_type_description(Term), Chars),
 	atom_chars(Atom, Chars).
 to_atom(Term, any) :-
 	pce_error(type_error(to_atom(Term, any), 1, ground, Term)).
