@@ -1,11 +1,9 @@
-/*  $Id$
-
-    Part of XPCE --- The SWI-Prolog GUI toolkit
+/*  Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (C): 1985-2011, University of Amsterdam
+    Copyright (C): 1985-2012, University of Amsterdam
 			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
@@ -33,34 +31,14 @@
 :- module(emacs_prolog_mode, []).
 :- use_module(library(pce)).
 :- use_module(library(debug)).
-:- use_module(library(edit)).
+:- use_module(library(make)).			% for reloading files
 :- use_module(library(emacs_extend)).
-:- use_module(library(error)).
 :- use_module(library(lists)).
-:- use_module(library(operators)).
 :- use_module(library(pce_meta)).
 :- use_module(library(pce_prolog_xref)).
-:- use_module(library(predicate_options)).
 :- use_module(library(prolog_colour)).
-:- use_module(library(prolog_predicate)).
+:- use_module(library(prolog_predicate)).	% class prolog_predicate
 :- use_module(library(prolog_source)).
-:- require([ make/0
-	   , absolute_file_name/3
-	   , auto_call/1
-	   , chain_list/2
-	   , atomic_list_concat/2
-	   , default/3
-	   , forall/2
-	   , ignore/1
-	   , last/2
-	   , list_to_set/2
-	   , member/2
-	   , memberchk/2
-	   , seek/4
-	   , strip_module/3
-	   ]).
-pce_ifhostproperty(prolog(quintus),
-		   (:- use_module(library(strings), [concat_chars/2]))).
 
 resource(mode_pl_icon, image, image('32x32/doc_pl.xpm')).
 resource(breakpoint,   image, image('16x16/stop.xpm')).
@@ -71,6 +49,7 @@ resource(breakpoint,   image, image('16x16/stop.xpm')).
 	[ insert_if_then_else	       = key('(') + key(';') + key('>'),
 	  insert_quote		       = key('"'),
 	  insert_percent	       = key('%'),
+	  insert_quasi_quote	       = key('|'),
 
 	  newline_and_indent	       = key('RET'),
 
@@ -100,6 +79,8 @@ resource(breakpoint,   image, image('16x16/stop.xpm')).
 	  consult_selection	       = button(compile) + button(compile),
 	  source_file		       = button(browse,
 						@prolog?source_file_chain),
+	  loaded_from		       = button(browse,
+						@emacs_mode?loaded_from_chain),
 
 	  forward_clause	       = key('\\ee'),
 	  backward_clause	       = key('\\ea'),
@@ -131,6 +112,7 @@ resource(breakpoint,   image, image('16x16/stop.xpm')).
 	  '\\n' + comment_end,
 	  '/'  + comment_start('*'),
 	  '*'  + comment_end('/'),
+	  quasi_quotation('||', '|}'),
 
 	  paragraph_end([ '\\s*$',		% empty line
 			  '/\\*',		% comment start
@@ -161,10 +143,13 @@ variable(warnings,	   int := 0,	 get, "Number of warnings").
 variable(errors,	   int := 0,	 get, "Number of errors").
 variable(body_indentation, int,		 get, "Indentation for body-goals").
 variable(cond_indentation, int,		 get, "Indent step for conditional").
+variable(quasiquotation_syntax,
+			   name*,	 both, "Default quasiquotation syntax").
 
-class_variable(body_indentation, int,  8).
-class_variable(cond_indentation, int,  4).
-class_variable(indent_tabs,      bool, @on,
+class_variable(quasiquotation_syntax, name*, @nil).
+class_variable(body_indentation,      int,   8).
+class_variable(cond_indentation,      int,   4).
+class_variable(indent_tabs,           bool,  @on,
 	       "Use tabs for indentation").
 
 
@@ -234,18 +219,40 @@ indent_line(E) :->
 	send(E, beginning_of_text_on_line),
 	get(E, caret, Caret),
 	get(E, beginning_of_clause, Caret, Base),
-	(   send(E, indent_comment_line)
-	;   send(E, indent_close_bracket_line, ')}]', Base)
-	;   send(E, indent_if_then_else)
-	;   send(E, indent_expression_line, ')}]', Base)
-	;   send(E, indent_clause_line)
-	;   get(E, body_indentation, Indent),
+	(   send(E, indent_comment_line),			W = comment
+	;   send(E, indent_close_bracket_line, ')}]', Base),	W = close_bracket
+	;   send(E, indent_if_then_else),			W = if_then_else
+	;   send(E, indent_expression_line, ')}]', Base),	W = expression
+	;   send(E, indent_clause_line),			W = clause
+	;   get(E, body_indentation, Indent),			W = body,
 	    send(E, align_line, Indent)
+	),
+	debug(emacs(indent), 'Indented line as ~w', [W]),
+	send(E, fixup_if_then_else).
+
+
+% ->fixup_if_then_else fixes changed alignment due to tabs
+% that expand to a different number of spaces.  Now only
+% does this for the first one.  Probably should do this
+% for the entire line.
+
+fixup_if_then_else(E) :->
+	"Fixup [(;->]\t"::
+	get(E, caret, C),
+	get(E, column, C, Indent),
+	(   get(E, looking_at, '(\\(|;|->)\t', Len)
+	->  get(E, caret, Caret),
+	    get(E, column, Caret+Len, Col),
+	    get(E, cond_indentation, Extra),
+	    (   Col =:= Indent+Extra
+	    ->  true
+	    ;   send(E, align, Indent+Extra, Caret+Len)
+	    )
+	;   true
 	).
 
-
 beginning_of_clause(E, Start:int, BOP:int) :<-
-	"Find start of predicate"::
+	"Find start of the a clause that contains Start"::
 	new(Here, number(Start)),
 	get(E, text_buffer, TB),
 	repeat,
@@ -272,18 +279,24 @@ beginning_of_clause(E) :->
 	get(E, beginning_of_clause, Caret, BOC),
 	send(E, caret, BOC).
 
-beginning_of_if_then_else(E, Pos:int) :<-
+beginning_of_if_then_else(E, New:[bool], OpenPos:int) :<-
 	"Beginning of if-then-else construct"::
 	get(E, caret, Caret),
 	get(E, text_buffer, TB),
 	pce_catch_error(mismatched_bracket,
-			get(TB, matching_bracket, Caret, ')', Pos)),
-	get(TB, character, Pos-1, Before),
+			get(TB, matching_bracket, Caret, ')', OpenPos)),
+	get(TB, character, OpenPos-1, Before),
 	\+ send(E?syntax, has_syntax, Before, word),
 	Before \== 0'?,				% '?(' for xpce
+	(   New \== @on
+	->  get(TB, scan, OpenPos, line, 0, end, EOL), % see <-argument_indent
+	    get(TB, skip_comment, OpenPos+1, EOL, P1),
+	    P1 \== EOL
+	;   true
+	),
 	get(E, beginning_of_clause, Caret, BegOfPred),
-	BegOfPred < Pos,
-	get(TB, scan_syntax, BegOfPred, Pos, tuple(code,_)).
+	BegOfPred < OpenPos,
+	get(TB, scan_syntax, BegOfPred, OpenPos, tuple(code,_)).
 
 
 indent_if_then_else(E) :->
@@ -333,7 +346,7 @@ insert_if_then_else(E, Times:[int], Char:char) :->
 	get(TB, scan, Caret, line, 0, start, SOL),
 	(   get(regex('\\s*(\\(|->|;)'), match, TB, SOL, L),
 	    Caret =:= SOL + L,
-	    get(E, beginning_of_if_then_else, OpenPos)
+	    get(E, beginning_of_if_then_else, @on, OpenPos)
 	->  get(E, column, OpenPos, Col),
 	    get(E, cond_indentation, Indent),
 	    send(E, align, Col+Indent)
@@ -423,6 +436,22 @@ indent_comment_line(E) :->
 	).
 
 
+insert_quasi_quote(E) :->
+	"Deal with {|Syntax||Quoted|}>"::
+	send(E, insert, '|'),
+	get(E, caret, Here),
+	(   send(E, looking_at, '{\\|', Here, 0)
+	->  (   get(E, quasiquotation_syntax, Syntax),
+	        Syntax \== @nil
+	    ->	send(E, insert, Syntax),
+		send(E, insert, '|||}'),
+		send(E, backward_char, 2)
+	    ;	send(E, insert, '|||}'),
+		send(E, backward_char, 4)
+	    )
+	;   true
+	).
+
 
 		 /*******************************
 		 *          COMPILATION		*
@@ -441,15 +470,30 @@ compile_buffer(E) :->
 	get(E?text_buffer, file, File),
 	(   send(File, instance_of, file)
 	->  send(E, save_if_modified),
-	    get(File, name, Path),
-	    print_message(silent, emacs(consult(user:Path))),
-	    consult(user:Path),
-	    print_message(silent, emacs(consulted(user:Path))),
-	    send(E, report, status, '%s compiled', Path)
+	    get(File, absolute_path, Path0),
+	    absolute_file_name(Path0, Path),
+	    master_load_file(Path, [], ToLoad),
+	    print_message(silent, emacs(consult(user:ToLoad))),
+	    make:reload_file(ToLoad),
+	    print_message(silent, emacs(consulted(user:ToLoad))),
+	    send(E, report, status, '%s compiled', ToLoad)
 	;   send(E, report, error,
 		 'Buffer is not connected to a file')
 	).
 
+%%	master_load_file(+File, +Seen, -MasterFile) is det.
+%
+%	If file is included into another  file, find the outermost file.
+%	This is the file that needs to  be reloaded instead of reloading
+%	File.
+
+master_load_file(File0, Seen, File) :-
+	source_file_property(File0, included_in(File1, _Line)),
+	\+ memberchk(File1, Seen), !,
+	master_load_file(File1, [File0|Seen], File).
+master_load_file(File, _, File).
+
+
 close_warning_window(_E) :->
 	"Destroy compilation error window"::
 	(   object(@prolog_warnings)
@@ -476,13 +520,18 @@ find_definition(M, For:prolog_predicate, Where:[{here,tab,window}]) :->
 	"Find definition of predicate [in new window]"::
 	get(M, text_buffer, TB),
 	get(For, head, @off, Head),
-	(   (   xref_defined(TB, Head, local(Line))		% local
-	    ;	xref_defined(TB, Head, constraint(Line))
-	    ;   xref_defined(TB, Head, foreign(Line))
+	(   (   xref_defined(TB, Head, local(Location))		% local
+	    ;	xref_defined(TB, Head, constraint(Location))
+	    ;   xref_defined(TB, Head, foreign(Location))
 	    )
 	->  get(TB, open, Where, Frame),
 	    get(Frame, editor, Editor),
-	    send(Editor, goto_line, Line, title := For?print_name)
+	    (	integer(Location)
+	    ->	send(Editor, goto_line, Location, title := For?print_name)
+	    ;	Location = (File:Line)
+	    ->  send(@emacs, goto_source_location,
+		     source_location(File, Line), tab)
+	    )
 	;   xref_defined(TB, Head, imported(File))	% imported
 	->  new(B, emacs_buffer(File)),
 	    get(B, open, Where, EmacsFrame),
@@ -504,16 +553,68 @@ find_local_definition(M, For:prolog_predicate) :->
 	"Find Prolog predicate in local buffer"::
 	get(M, text_buffer, TB),
 	get(For, head, @off, Head),
-	(   (   xref_defined(TB, Head, local(Line))
+	(   (   xref_defined(TB, Head, local(Location))
 	    ->  true
 	    ;   send(M, xref_buffer),
-		xref_defined(TB, Head, local(Line))
+		get(M, xref_source_id, SourceID),
+		xref_defined(SourceID, Head, local(Location))
+	    )
+	->  (   integer(Location)
+	    ->	send(M, goto_line, Location, title := For?print_name)
+	    ;	Location = (File:Line)
+	    ->	send(@emacs, goto_source_location,
+		     source_location(File, Line), tab)
 	    )
-	->  send(M, goto_line, Line, title := For?print_name)
 	;   send(M, report, warning, 'Cannot find %N', For)
 	).
 
 
+		 /*******************************
+		 *	   LOAD CONTEXT		*
+		 *******************************/
+
+loaded_from(_M, LoadedFrom:source_location) :->
+	"Jump to position I'm loaded from"::
+	send(@emacs, goto_source_location, LoadedFrom, tab).
+
+loaded_from_chain(M, LoadedFrom:chain) :<-
+	"Chain with files and locations I'm loaded from"::
+	get(M, file, File), File \== @nil,
+	get(File, absolute_path, Path0),
+	absolute_file_name(Path0, Path),
+	findall(SLoc, loaded_from(Path, SLoc), Locations),
+	Term =.. [chain|Locations],
+	new(LoadedFrom, Term).
+
+loaded_from(Path, source_location(File, Line)) :-
+	(   source_file_property(Path, load_context(_Module, File:Line, _Options))
+	;   source_file_property(Path, included_in(File, Line))
+	).
+
+
+		 /*******************************
+		 *	       INFO		*
+		 *******************************/
+
+file_module(M, Module:name) :<-
+	"Module used for the file"::
+	get(M, text_buffer, TB),
+	(   xref_module(TB, Module)
+	->  true
+	;   get(TB, file, File), File \== @nil,
+	    get(File, absolute_path, Path0),
+	    absolute_file_name(Path0, Path),
+	    module_context(Path, [], Module)
+	).
+
+properties(M, V:view) :<-
+	get_super(M, properties, V),
+	(   get(M, file_module, Module)
+	->  send(V, appendf, 'Prolog module:\t%s\n', Module)
+	;   true
+	).
+
+
 		 /*******************************
 		 *	    PCE CLASSES		*
 		 *******************************/
@@ -581,14 +682,14 @@ pce_define_class(M, Name:name,
 	get(Comment, value, C),
 	(   C == ''
 	->  send(M, format,
-		 ':- pce_begin_class(%s, %s).\n\n\
+		 ':- pce_begin_class(%s, %s).\n\n\c
 		  :- pce_end_class(%s).\n',
 		 QName, QSuper, QName)
 	;   atom_codes(C, Codes),
-	    string_to_list(S, Codes),
+	    string_codes(S, Codes),
 	    format(string(QComment), '~q', S),
 	    send(M, format,
-		 ':- pce_begin_class(%s, %s, %s).\n\n\
+		 ':- pce_begin_class(%s, %s, %s).\n\n\c
 		  :- pce_end_class(%s).\n',
 		 QName, QSuper, QComment, QName)
 	),
@@ -746,11 +847,13 @@ do_spy(variable(Name, _Type, _Access, _Doc), M, (Class-Name)) :-
 do_spy((Head :- _Body), M, Spec) :-
 	prolog_debug_spec(M, Head, Spec),
 	user:spy(Spec).
+do_spy((Head --> _Body), M, Spec) :-
+	dcg_debug_spec(M, Head, Spec),
+	user:spy(Spec).
 do_spy(Head, M, Spec) :-
 	prolog_debug_spec(M, Head, Spec),
 	user:spy(Spec).
 
-
 trace(M) :->
 	"Set trace-point on implementation"::
 	get(M, prolog_term, Term),
@@ -788,6 +891,13 @@ prolog_debug_spec(M, Head, Spec) :-
 	(   get(M, prolog_module, Module)
 	->  Spec = (Module:Name/Arity)
 	;   Spec = Name/Arity
+	).
+
+dcg_debug_spec(M, Head, Spec) :-
+	catch(functor(Head, Name, Arity), _, fail),
+	(   get(M, prolog_module, Module)
+	->  Spec = (Module:Name//Arity)
+	;   Spec = Name//Arity
 	).
 
 		 /*******************************
@@ -826,12 +936,14 @@ error_at_location(M, Caret:int) :->
 	send(M, check_clause).
 
 
-symbol_chars("-#$&*+./:<=>?@\\^`~").
+symbol_chars('-#$&*+./:<=>?@\\^~').
 
+term_expansion(symbol_char(_), Clauses) :-
+	symbol_chars(String),
+	atom_codes(String, Codes),
+	findall(symbol_char(C), member(C,Codes), Clauses).
 
-symbol_char(C) :-
-	symbol_chars(Symbols),
-	memberchk(C, Symbols).
+symbol_char(_Code).
 
 
 insert_full_stop(M, Arg:[int]) :->
@@ -858,10 +970,8 @@ check_clause(M, From:from=[int], Repair:repair=[bool], End:int) :<-
         (   From == @default
 	->  get(M, caret, C),
 	    get(M, beginning_of_clause, C, Start),
-	    ignore(send(M, electric_caret, Start)),
-	    Verbose = true
-	;   Start = From,
-	    Verbose = false
+	    ignore(send(M, electric_caret, Start))
+	;   Start = From
 	),
 	get(M, colourise_clause, Start, TermPos),
 	(   TermPos = error_position(_StartClause, _EndClause, ErrorPos)
@@ -981,9 +1091,12 @@ prolog_term(M, From:[int], Silent:[bool], TermPos:[prolog], Clause:prolog) :<-
 	  ;   Start = From
 	  ),
 	  get(M, text_buffer, TB),
-	  pce_open(TB, read, Fd),
-	  read_term_from_stream(TB, Fd, Start, Clause, Error, _S, P, _C),
-	  close(Fd),
+	  setup_call_cleanup(
+	      pce_open(TB, read, Fd),
+	      ( set_stream_file(TB, Fd),
+		read_term_from_stream(TB, Fd, Start, Clause, Error, _S, P, _C)
+	      ),
+	      close(Fd)),
 	  ignore(P = TermPos),
 	  (   var(Error)
 	  ->  true
@@ -1294,7 +1407,7 @@ backward_predicate(M, P0:int, BPred:int) :<-
 		    fail
 		)
 	;   get(M, backward_clause, P0-1, P2),
-	    (	get(M, name_and_arity, P2, tuple(Name, Arity))
+	    (	get(M, name_and_arity, P2, _)
 	    ->	get(M, backward_predicate, P2+1, BPred)
 	    ;	BPred = P2
 	    )
@@ -1350,15 +1463,26 @@ colourise_clause(M, From:from=[int], TermPos:prolog) :<-
 	;   Start = From
 	),
 	get(M, text_buffer, TB),
+	get(M, xref_source_id, SourceID),
 	setup_call_cleanup(
 	    pce_open(TB, read, Stream),
-	    ( seek(Stream, Start, bof, _),
-	      prolog_colourise_term(Stream, TB, colour_item(M),
-				    [ subterm_positions(TermPos)
-				    ])
+	    ( set_stream_file(TB, Stream),
+	      seek(Stream, Start, bof, _),
+	      prolog_colourise_term(
+		  Stream, SourceID, colour_item(M),
+		  [ subterm_positions(TermPos)
+		  ])
 	    ),
 	    close(Stream)).
 
+set_stream_file(TB, Stream) :-
+	get(TB, file, File), File \== @nil, !,
+	get(File, absolute_path, Path0),
+	absolute_file_name(Path0, Path),	% Make sure it is canonical
+	set_stream(Stream, file_name(Path)).
+set_stream_file(_, _).
+
+
 :- dynamic
 	style_name/2.
 
@@ -1467,11 +1591,39 @@ xref_buffer(M, Always:[bool]) :->
 	    ;   get(TB, xref_generation, GRef),
 		GRef \== G
 	    )
-	->  send(M, report, progress, 'Cross-referencing buffer ...'),
-	    xref_source(TB),
+	->  xref_source_id(TB, SourceId),
+	    (	TB == SourceId
+	    ->	send(M, report, progress, 'Cross-referencing buffer ...')
+	    ;	send(M, report, progress, 'Cross-referencing %s ...', SourceId),
+	        send(TB, attribute, xref_source_id, SourceId)
+	    ),
+	    xref_source(SourceId, [silent(true)]),
 	    send(TB, xref_generation, G),
 	    send(M, report, done)
 	;   true
+	).
+
+xref_source_id(M, SourceId:any) :<-
+	"Xref source identifier"::
+	get(M, text_buffer, TB),
+	(   get(TB, attribute, xref_source_id, SourceId)
+	->  true
+	;   SourceId = TB
+	).
+
+%%	xref_source_id(+TextBuffer, -SourceID) is det.
+%
+%	Find the object we need  to   examine  for cross-referencing. If
+%	this is an included file, this is the corresponding main file.
+
+xref_source_id(TB, SourceId) :-
+	get(TB, file, File), File \== @nil,
+	get(File, absolute_path, Path0),
+	absolute_file_name(Path0, Path),
+	master_load_file(Path, [], Master),
+	(   Master == Path
+	->  SourceId = TB
+	;   SourceId = Master
 	).
 
 
@@ -1481,9 +1633,12 @@ xref_buffer(M, Always:[bool]) :->
 
 colourise_buffer(M) :-
 	get(M, text_buffer, TB),
+	get(M, xref_source_id, SourceID),
 	setup_call_cleanup(
 	    pce_open(TB, read, Stream),
-	    prolog_colourise_stream(Stream, TB, colour_item(M)),
+	    ( set_stream_file(TB, Stream),
+	      prolog_colourise_stream(Stream, SourceID, colour_item(M))
+	    ),
 	    close(Stream)).
 
 
@@ -1498,13 +1653,15 @@ make_fragment(goal(Class, Goal), M, F, L, Style) :-
 	functor(Goal, Name, Arity),
 	send(Fragment, name, Name),
 	send(Fragment, arity, Arity),
-	(   Class =.. [ClassName,Context],
-	    atomic(Context)
-	->  send(Fragment, classification, ClassName),
-	    send(Fragment, context, Context)
-	;   functor(Class, ClassName, _),
-	    send(Fragment, classification, ClassName)
-	).
+	set_xref_fragment_context(Fragment, Class).
+make_fragment(head(Class, Head), M, F, L, Style) :-
+	callable(Head), !,
+	get(M, text_buffer, TB),
+	new(Fragment, emacs_head_fragment(TB, F, L, Style)),
+	functor(Head, Name, Arity),
+	send(Fragment, name, Name),
+	send(Fragment, arity, Arity),
+	set_xref_fragment_context(Fragment, Class).
 make_fragment(class(Type, Class), M, F, L, Style) :-
 	atom(Class), !,
 	get(M, text_buffer, TB),
@@ -1532,11 +1689,27 @@ make_fragment(Class, M, F, L, Style) :-
 make_simple_fragment(Class, M, F, L, Style) :-
 	get(M, text_buffer, TB),
 	new(Fragment, emacs_prolog_fragment(TB, F, L, Style)),
+	functor(Class, Classification, Arity),
+	send(Fragment, classification, Classification),
+	(   Arity >= 1,
+	    arg(1, Class, Context),
+	    atomic(Context)
+	->  send(Fragment, context, Context)
+	;   true
+	).
+
+set_xref_fragment_context(Fragment, Class) :-
 	functor(Class, Classification, Arity),
 	send(Fragment, classification, Classification),
 	(   Arity == 1
 	->  arg(1, Class, Context),
-	    send(Fragment, context, Context)
+	    (   atomic(Context)
+	    ->	send(Fragment, context, Context)
+	    ;	ground(Context),
+		Context = (Include:Line)
+	    ->	send(Fragment, context, source_location(Include, Line))
+	    ;	true
+	    )
 	;   true
 	).
 
@@ -1564,7 +1737,7 @@ popup(_GF, Popup:popup) :<-
 :- pce_global(@prolog_mode_goal_popup,
 	      make_prolog_mode_goal_popup).
 
-%	make_prolog_mode_goal_popup(-Popup)
+%%	make_prolog_mode_goal_popup(-Popup)
 %
 %	Create the popup and define actions for handling the right-menu
 %	on goals.
@@ -1572,8 +1745,9 @@ popup(_GF, Popup:popup) :<-
 make_prolog_mode_goal_popup(G) :-
 	new(G, popup(goal_actions)),
 	Fragment = @arg1,
-	new(HasSource, message(Fragment, has_source)),
+	new(HasSource,  message(Fragment, has_source)),
 	new(HasListing, message(Fragment, has_listing)),
+	new(HasInfo,    message(Fragment, has_info)),
 	send_list(G, append,
 		  [ menu_item(edit_in_tab,
 			      message(Fragment, edit, tab),
@@ -1585,6 +1759,9 @@ make_prolog_mode_goal_popup(G) :-
 			      message(Fragment, edit, window),
 			      condition := HasSource),
 		    gap,
+		    menu_item(info,
+			      message(Fragment, info),
+			      condition := HasInfo),
 		    menu_item(listing,
 			      message(Fragment, listing),
 			      condition := HasListing),
@@ -1596,48 +1773,79 @@ make_prolog_mode_goal_popup(G) :-
 
 module(F, Module:name) :<-
 	"Module for Module:Goal references"::
-	get(F, classification, extern),
-	get(F, context, Module),
-	Module \== @nil.
+	(   get(F, classification, extern),
+	    get(F, context, Module),
+	    Module \== @nil
+	->  true
+	;   get(F, file_module, Module)
+	).
+
+
+file_module(F, Module:name) :<-
+	"Module used for the file"::
+	get(F, text_buffer, TB),
+	(   xref_module(TB, Module)
+	->  true
+	;   get(TB, file, File), File \== @nil,
+	    get(File, absolute_path, Path0),
+	    absolute_file_name(Path0, Path),
+	    module_context(Path, [], Module)
+	).
+
+
+module_context(File, _, Module) :-
+	source_file_property(File, module(Module)), !.
+module_context(File, Seen, Module) :-
+	source_file_property(File, included_in(File2, _Line)),
+	\+ memberchk(File, Seen), !,
+	module_context(File2, [File|Seen], Module).
+module_context(File, _, Module) :-
+	source_file_property(File, load_context(Module, _, _)).
+
 
 predicate(F, Pred:prolog_predicate) :<-
 	"Get referenced predicate"::
 	get(F, name, Name),
 	get(F, arity, Arity),
 	(   get(F, module, Module)
-	->  Spec = Name/Arity
-	;   Spec = Module:Name/Arity
+	->  Spec = Module:Name/Arity
+	;   Spec = Name/Arity
 	),
 	new(Pred, prolog_predicate(Spec)).
 
-head(F, Head:prolog) :<-
+head(F, Qualify:[bool], Head:prolog) :<-
 	"Return goal-head"::
 	get(F, name, Name),
 	get(F, arity, Arity),
 	functor(Head0, Name, Arity),
-	(   get(F, module, M)
-	->  Head = M:Head0
-	;   Head = Head0
+	(   Qualify == @off			% @off: only if qualification is
+	->  (   get(F, classification, extern),  % explicit.
+	        get(F, context, Module),
+		Module \== @nil
+	    ->  Head = Module:Head0
+	    ;	Head = Head0
+	    )
+	;   (   get(F, module, M)
+	    ->  Head = M:Head0
+	    ;   Head = Head0
+	    )
 	).
 
-loaded_specifier(F, TheHead:prolog) :<-
+loaded_specifier(F, Head:prolog) :<-
 	"Get predicate specifier for loaded predicate"::
 	get(F, head, Head),
-	(   Head = _:_
-	->  TheHead = Head
-	;   get(F, text_buffer, TB),
-	    xref_module(TB, M)
-	->  TheHead = M:Head
-	;   TheHead = _:Head
-	),
-	current_predicate(_, TheHead).
+	current_predicate(_, Head).
 
 has_source(F) :->
 	"Test if there is source available"::
 	get(F, text_buffer, TB),
-	get(F, head, Head),
+	get(F, head, @off, Head),
 	(   xref_defined(TB, Head, How),
 	    xref_definition_line(How, _)
+	;   Head = Module:Plain,
+	    xref_module(Src, Module),
+	    xref_defined(Src, Plain, How),
+	    xref_definition_line(How, _)
 	;   xref_defined(TB, Head, imported(_From))
 	;   get(prolog_predicate(Head), source, _)
 	), !.
@@ -1661,13 +1869,12 @@ listing(F) :->
 	"Generate a listing"::
 	get(F, loaded_specifier, Spec),
 	new(Tmp, emacs_buffer(@nil, string('*Listing for %N*', F))),
-	pce_open(Tmp, write, Out),
-	telling(Old), set_output(Out),
-	ignore(listing(Spec)),
-	tell(Old),
-	close(Out),
+	send(Tmp, mode, prolog),
+	setup_call_cleanup(
+	    pce_open(Tmp, write, Out),
+	    with_output_to(Out, listing(Spec)),
+	    close(Out)),
 	send(Tmp, modified, @off),
-%	send(Tmp, mode, prolog),
 	send(Tmp, open, tab).
 
 
@@ -1677,6 +1884,16 @@ has_listing(F) :->
 	predicate_property(Spec, number_of_clauses(N)),
 	N > 0.
 
+info(F) :->
+	"Provide all know information about P"::
+	get(F, predicate, P),
+	send(P, info).
+
+has_info(F) :->
+	"Provide all know information about P"::
+	get(F, predicate, P),
+	get(P, head, Head),
+	predicate_property(Head, _), !.
 
 documentation(F) :->
 	"Invoke Prolog help-system"::
@@ -1738,6 +1955,10 @@ identify_pred(autoload, F, Summary) :-	% Autoloaded predicates
 	new(Summary, string('%N: autoload from %s', F, Atom)).
 identify_pred(local(Line), F, Summary) :-	% Local predicates
 	new(Summary, string('%N: locally defined at line %d', F, Line)).
+identify_pred(include(SrcLoc), F, Summary) :-
+	get(SrcLoc, file_name, File),
+	get(SrcLoc, line_no, Line),
+	new(Summary, string('%N: included from %s:%d', F, File, Line)).
 identify_pred(public(Line), F, Summary) :-	% Public predicates
 	new(Summary, string('%N: declared public at line %d', F, Line)).
 identify_pred(foreign(Line), F, Summary) :-	% Foreign predicates
@@ -1764,7 +1985,8 @@ identify_pred(Class, _, ClassName) :-
 autoload_source(F, Source) :-
 	get(F, name, Name),
 	get(F, arity, Arity),
-	'$find_library'(_Module, Name, Arity, _LoadModule, Library), !,
+	functor(Head, Name, Arity),
+	predicate_property(Head, autoload(Library)),
 	absolute_file_name(Library, Source,
 			   [ file_type(prolog),
 			     access(read),
@@ -1774,6 +1996,108 @@ autoload_source(F, Source) :-
 :- pce_end_class(emacs_goal_fragment).
 
 
+:- pce_begin_class(emacs_head_fragment, emacs_goal_fragment,
+		   "Fragment for a predicate head in PceEmacs").
+
+:- pce_group(popup).
+
+popup(_GF, Popup:popup) :<-
+	"Return popup menu"::
+	Popup = @prolog_mode_head_popup.
+
+:- pce_global(@prolog_mode_head_popup,
+	      make_prolog_mode_head_popup).
+
+%%	make_prolog_mode_head_popup(-Popup)
+%
+%	Create the popup and define actions for handling the right-menu
+%	on predicate heads.
+
+make_prolog_mode_head_popup(G) :-
+	new(G, popup(head_actions)),
+	Fragment = @arg1,
+	new(HasListing, message(Fragment, has_listing)),
+	new(HasInfo,    message(Fragment, has_info)),
+	send_list(G, append,
+		  [ menu_item(info,
+			      message(Fragment, info),
+			      condition := HasInfo),
+		    menu_item(listing,
+			      message(Fragment, listing),
+			      condition := HasListing),
+		    gap,
+		    menu_item(documentation,
+			      message(Fragment, documentation))
+		  ]).
+
+identify(F) :->
+	"Tell the user about the predicate"::
+	get(F, text_buffer, TB),
+	get(F, classification, Class),
+	(   get(F, context, Context),
+	    Context \== @nil
+	->  Id =.. [Class, Context]
+	;   Id = Class
+	),
+	identify_head(Id, F, Report),
+	send(TB, report, status, Report).
+
+identify_head(Class, F, Summary) :-
+	get(F, print_name, Name), !,
+	(   get(F, loaded_specifier, Head)
+	->  findall(Prop, head_property(Head, Prop), Props),
+	    atomic_list_concat(Props, ', ', Text),
+	    new(Summary, string('%s: (loaded) %s', Name, Text))
+	;   term_to_atom(Class, Text),
+	    new(Summary, string('%s: (not loaded) %s', Name, Text))
+	).
+
+head_property(Head, Text) :-
+	predicate_property(Head, Prop),
+	\+ hidden_property(Prop, Head),
+	(   atomic(Text)
+	->  Text = Prop
+	;   property_text(Prop, Text)
+	).
+
+hidden_property(file(_), _).
+hidden_property(line_count(_), _).
+hidden_property(nodebug, _).
+hidden_property(interpreted, _).
+hidden_property(visible, _).
+hidden_property(transparent, Head) :-
+	predicate_property(Head, meta_predicate(_)).
+
+property_text(number_of_clauses(N), Text) :- !,
+	(   N == 1
+	->  Text = '1 clause'
+	;   atomic_list_concat([N, ' clauses'], Text)
+	).
+property_text(indexed(List), Text) :- !,
+	(   List = [1-_]
+	->  Text = 'hashed on first argument'
+	;   List = [N-_]
+	->  int_postfix(N, PostFix),
+	    format(atom(Text), 'hashed on ~d-~w argument', [N, PostFix])
+	;   pairs_keys(List, Args),
+	    atomic_list_concat(Args, ', ', ArgText),
+	    format(atom(Text), 'hashed on arguments ~w', [ArgText])
+	).
+property_text(meta_predicate(Head), Text) :- !,
+	Head =.. [_|Args],
+	Meta =.. [meta_predicate|Args],
+	term_to_atom(Meta, Text).
+property_text(Term, Text) :-
+	term_to_atom(Term, Text).
+
+int_postfix(1, st) :- !.
+int_postfix(2, nd) :- !.
+int_postfix(3, rd) :- !.
+int_postfix(_, th).
+
+:- pce_end_class(emacs_head_fragment).
+
+
 		 /*******************************
 		 *	  CLASS FRAGMENT	*
 		 *******************************/
@@ -1923,16 +2247,20 @@ variable(context,	 prolog*, both,	"Classification argument").
 popup(F, Popup:popup) :<-
 	get(F, context, Context),
 	Context \== @nil,
-	(   get(F, classification, file)
-	;   get(F, classification, module)
-	),
-	Popup = @prolog_mode_file_popup.
+	get(F, classification, FragClass),
+	fragment_popup(FragClass, Popup).
+
+fragment_popup(file,	       @prolog_mode_file_popup).
+fragment_popup(file_no_depend, @prolog_mode_file_popup).
+fragment_popup(module,	       @prolog_mode_module_popup).
 
 :- pce_global(@prolog_mode_file_popup,
 	      make_prolog_mode_file_popup).
+:- pce_global(@prolog_mode_module_popup,
+	      make_prolog_mode_module_popup).
 
-make_prolog_mode_file_popup(G) :-
-	new(G, popup(file_actions)),
+make_prolog_mode_module_popup(G) :-
+	new(G, popup(actions)),
 	send_list(G, append,
 		  [ menu_item(open_in_tab,
 			      message(@emacs, open_file, @arg1?file, tab)),
@@ -1942,15 +2270,68 @@ make_prolog_mode_file_popup(G) :-
 			      message(@emacs, open_file, @arg1?file, here))
 		  ]).
 
+make_prolog_mode_file_popup(G) :-
+	make_prolog_mode_module_popup(G),
+	send_list(G, append,
+		  [ gap,
+		    new(R, popup(resolves, message(@arg1, edit)))
+		  ]),
+	send(R, update_message,
+	     message(@arg1, add_resolve_items, R)).
+
+
 file(F, File:name) :<-
 	"Return associated file"::
 	get(F, context, Context),
-	(   get(F, classification, file)
+	(   (   get(F, classification, file)
+	    ;	get(F, classification, file_no_depend)
+	    )
 	->  File = Context
 	;   get(F, classification, module)
 	->  module_property(Context, file(File))
 	).
 
+add_resolve_items(F, Popup:popup) :->
+	"Add entries for resolved predicates"::
+	send(Popup, clear),
+	get(F, context, File),
+	get(F, text_buffer, TB),
+	(   get(TB, attribute, xref_source_id, SourceId)
+	->  true
+	;   SourceId = TB
+	),
+	findall(Head, ( xref_defined(SourceId, Head, imported(File)),
+			xref_called(SourceId, Head, _)
+		      ),
+		UsedHeads0),
+	(   UsedHeads0 == []
+	->  (   \+ xref_defined(SourceId, Head, imported(File))
+	    ->	Comment = 'File has no exports'
+	    ;	Comment	= 'File resolves no predicates'
+	    ),
+	    send(Popup, append, menu_item(Comment, @nil))
+	;   (   source_file_property(File, module(FromModule))
+	    ->  maplist(head_pi(FromModule), UsedHeads0, UsedPreds0)
+	    ;   maplist(head_pi, UsedHeads0, UsedPreds0)
+	    ),
+	    sort(UsedPreds0, UsedPreds),
+	    forall(member(Pred, UsedPreds),
+		   (   new(PredObj, prolog_predicate(Pred)),
+		       send(Popup, append,
+			    menu_item(PredObj, @default, PredObj?print_name))
+		   ))
+	).
+
+head_pi(M0, Head, M:Name/Arity) :-
+	strip_module(M0:Head, M, Term),
+	functor(Term, Name, Arity).
+
+head_pi(M0:Head, M:Name/Arity) :- !,
+	strip_module(M0:Head, M, Term),
+	functor(Term, Name, Arity).
+head_pi(Term, Name/Arity) :- !,
+	functor(Term, Name, Arity).
+
 
 identify(F) :->
 	"Identify in status window"::
@@ -1963,9 +2344,17 @@ identify(F) :->
 	identify_fragment(Term, F, Summary), !,
 	send(F, report, status, Summary).
 
+:- multifile
+	identify/2.
+
+identify_fragment(Term, _F, Text) :-
+	phrase(syntax_message(Term), List), !,
+	elements_to_string(List, Text).
 identify_fragment(var,  _, 'Variable').
 identify_fragment(file(Path), _, Summary) :-
 	new(Summary, string('File %s', Path)).
+identify_fragment(file_no_depend(Path), _, Summary) :-
+	new(Summary, string('File %s (does not resolve any predicates)', Path)).
 identify_fragment(directory(Path), _, Summary) :-
 	new(Summary, string('Directory %s', Path)).
 identify_fragment(type_error(Type), _, Summary) :-
@@ -1979,12 +2368,23 @@ identify_fragment(method(send), _, 'XPCE send method').
 identify_fragment(method(get), _, 'XPCE get method').
 identify_fragment(head(unreferenced), _, 'Unreferenced predicate (from this file)').
 identify_fragment(head(exported), _, 'Exported predicate').
-identify_fragment(head(public), _, 'Public predicate').
+identify_fragment(head(public(_)), _,
+		  'Public predicate (not exported, but called from elsewhere)').
 identify_fragment(head(multifile), _, 'Multifile predicate').
-identify_fragment(head(constraint), _, 'Constraint').
+identify_fragment(head(constraint), _, 'Constraint (CHR)').
+identify_fragment(head(iso), _, 'ISO built-in (cannot be redefined)').
+identify_fragment(head(built_in), _, 'SWI-Prolog built-in (cannot be redefined)').
+identify_fragment(head(def_iso), _, 'Definition of an ISO built-in').
+identify_fragment(head(def_swi), _, 'Definition of an SWI-Prolog built-in').
+identify_fragment(head(imported(From)), _, Summary) :-
+	new(Summary, string('Also imported from %s', From)).
 identify_fragment(prolog_data, _, 'Pass Prolog term unmodified').
 identify_fragment(keyword(except), _, 'Import all except given').
 identify_fragment(keyword(as), _, 'Import under a different name').
+identify_fragment(unused_import, _, 'Imported predicate is not used').
+identify_fragment(undefined_import, _, 'Predicate is not exported').
+identify_fragment(dcg_right_hand_ctx, _, 'right-hand-context (DCG `push-back\')').
+identify_fragment(goal(not_callable), _, 'Goal is not callable (type error)').
 identify_fragment(Class, _, Summary) :-
 	term_to_atom(Class, Summary).
 
@@ -2004,4 +2404,13 @@ add_one_ofs([H|T], Summary) :-
 	    add_one_ofs(T, Summary)
 	).
 
+elements_to_string(List, String) :-
+	maplist(element_to_string, List, Parts),
+	atomic_list_concat(Parts, String).
+
+element_to_string(Fmt-Args, String) :- !,
+	format(string(String), Fmt, Args).
+element_to_string(nl, '\n') :- !.
+element_to_string(Atom, Atom).
+
 :- pce_end_class(emacs_prolog_fragment).
