@@ -615,7 +615,7 @@ markAtom(atom_t a)
 
   ap = fetchAtomArray(i);
 
-  if ( ap )
+  if ( ap && !(ap->references & ATOM_MARKED_REFERENCE) )
   {
 #ifdef O_DEBUG_ATOMGC
     if ( atomLogFd )
@@ -629,6 +629,35 @@ markAtom(atom_t a)
   }
 }
 
+void
+unmarkAtoms(void)
+{ size_t index;
+  int i, last=FALSE;
+
+  for(index=GD->atoms.builtin, i=MSB(index); !last; i++)
+  { size_t upto = (size_t)2<<i;
+    Atom *b = GD->atoms.array.blocks[i];
+
+    if ( upto >= GD->atoms.highest )
+    { upto = GD->atoms.highest;
+      last = TRUE;
+    }
+
+    for(; index<upto; index++)
+    { Atom a = b[index];
+
+      if ( a && (a->references & ATOM_MARKED_REFERENCE) )
+      {
+#ifdef ATOMIC_REFERENCES
+        ATOMIC_AND(&a->references, ~ATOM_MARKED_REFERENCE);
+#else
+        a->references &= ~ATOM_MARKED_REFERENCE;
+#endif
+      }
+    }
+  }
+}
+
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 destroyAtom()  actually  discards  an  atom.  The  code  marked  (*)  is
@@ -794,9 +823,10 @@ pl_garbage_collect_atoms(void)
   GD->atoms.gc_active = TRUE;
   blockSignals(&set);
   t = CpuTime(CPU_USER);
+  unmarkAtoms();
   markAtomsOnStacks(LD);
 #ifdef O_PLMT
-  forThreadLocalData(markAtomsOnStacks, 0);
+  forThreadLocalDataUnsuspended(markAtomsOnStacks, 0);
   markAtomsMessageQueues();
 #endif
   oldcollected = GD->atoms.collected;
@@ -1058,13 +1088,15 @@ registerBuiltinAtoms(void)
 
 
 #if O_DEBUG
-static void
+static int
 exitAtoms(int status, void *context)
 { (void)status;
   (void)context;
 
   Sdprintf("hashstat: %d lookupAtom() calls used %d strcmp() calls\n",
 	   lookups, cmps);
+
+  return 0;
 }
 #endif
 
@@ -1284,19 +1316,56 @@ PRED_IMPL("$atom_references", 2, atom_references, 0)
 #define ALT_MAX 256		/* maximum number of alternatives */
 #define stringMatch(m)	((m)->name->name)
 
+#define is_signalled() unlikely(LD && LD->signal.pending != 0)
+
 typedef struct match
 { Atom		name;
   size_t	length;
 } *Match;
 
 
-static bool
-allAlpha(const char *s, size_t len)
-{ for( ; len-- > 0; s++)
-  { if ( !*s || !isAlpha(*s) )
-      fail;
+static inline int
+completion_candidate(Atom a)
+{ return (a->references || indexAtom(a->atom) < GD->atoms.builtin);
+}
+
+
+static int
+is_identifier_text(PL_chars_t *txt)
+{ if ( txt->length == 0 )
+    return FALSE;
+
+  switch(txt->encoding)
+  { case ENC_ISO_LATIN_1:
+    { const unsigned char *s = (const unsigned char *)txt->text.t;
+      const unsigned char *e = &s[txt->length];
+
+      if ( !f_is_prolog_atom_start(*s) )
+	return FALSE;
+
+      for(s++; s<e; s++)
+      { if ( !f_is_prolog_identifier_continue(*s) )
+	  return FALSE;
+      }
+      return TRUE;
+    }
+    case ENC_WCHAR:
+    { const pl_wchar_t *s = (const pl_wchar_t*)txt->text.w;
+      const pl_wchar_t *e = &s[txt->length];
+
+      if ( !f_is_prolog_atom_start(*s) )
+	return FALSE;
+
+      for(s++; s<e; s++)
+      { if ( !f_is_prolog_identifier_continue(*s) )
+	  return FALSE;
+      }
+      return TRUE;
+    }
+    default:
+      assert(0);
+      return FALSE;
   }
-  succeed;
 }
 
 
@@ -1322,6 +1391,7 @@ extendAtom(char *prefix, bool *unique, char *common)
     { Atom a = b[index];
 
       if ( a && a->type == &text_atom &&
+	   completion_candidate(a) &&
 	   strprefix(a->name, prefix) &&
 	   strlen(a->name) < LINESIZ )
       { if ( first == TRUE )
@@ -1343,9 +1413,27 @@ extendAtom(char *prefix, bool *unique, char *common)
 }
 
 
-word
-pl_complete_atom(term_t prefix, term_t common, term_t unique)
-{ char *p;
+/** '$complete_atom'(+Prefix, -Common, -Unique) is semidet.
+
+True when Prefix can be extended based on currently defined atoms.
+
+@arg Common is a code list consisting of the characters from Prefix
+     and the common text for all possible completions
+@arg Unique is either =unique= or =not_unique=.  In the second case,
+     this implies that there are longer atoms that have the prefix
+     Common.
+@see '$atom_completions'/2.
+@bug This version only handles ISO Latin 1 text
+*/
+
+static
+PRED_IMPL("$complete_atom", 3, complete_atom, 0)
+{ PRED_LD
+  term_t prefix = A1;
+  term_t common = A2;
+  term_t unique = A3;
+
+  char *p;
   bool u;
   char buf[LINESIZ];
   char cmm[LINESIZ];
@@ -1355,9 +1443,7 @@ pl_complete_atom(term_t prefix, term_t common, term_t unique)
   strcpy(buf, p);
 
   if ( extendAtom(p, &u, cmm) )
-  { GET_LD
-
-    strcat(buf, cmm);
+  { strcat(buf, cmm);
     if ( PL_unify_list_codes(common, buf) &&
 	 PL_unify_atom(unique, u ? ATOM_unique
 				 : ATOM_not_unique) )
@@ -1374,8 +1460,8 @@ compareMatch(const void *m1, const void *m2)
 }
 
 
-static bool
-extend_alternatives(char *prefix, struct match *altv, int *altn)
+static int
+extend_alternatives(PL_chars_t *prefix, struct match *altv, int *altn)
 { size_t index;
   int i, last=FALSE;
 
@@ -1391,11 +1477,16 @@ extend_alternatives(char *prefix, struct match *altv, int *altn)
 
     for(; index<upto; index++)
     { Atom a = b[index];
+      PL_chars_t hit;
 
-      if ( a && a->type == &text_atom &&
-	   strprefix(a->name, prefix) &&
-	   a->length < ALT_SIZ &&
-	   allAlpha(a->name, a->length) )
+      if ( index % 256 == 0 && PL_handle_signals() < 0 )
+	return FALSE;			/* interrupted */
+
+      if ( a && completion_candidate(a) &&
+	   get_atom_ptr_text(a, &hit) &&
+	   hit.length < ALT_SIZ &&
+	   PL_cmp_text(prefix, 0, &hit, 0, prefix->length) == 0 &&
+	   is_identifier_text(&hit) )
       { Match m = &altv[(*altn)++];
 
 	m->name = a;
@@ -1409,31 +1500,44 @@ extend_alternatives(char *prefix, struct match *altv, int *altn)
 out:
   qsort(altv, *altn, sizeof(struct match), compareMatch);
 
-  succeed;
+  return TRUE;
 }
 
 
-word
-pl_atom_completions(term_t prefix, term_t alternatives)
-{ GET_LD
-  char *p;
-  char buf[LINESIZ];
+/** '$atom_completions'(+Prefix, -Alternatives:list(atom)) is det.
+
+True when Alternatives is a list of   all  atoms that have prefix Prefix
+and are considered completion  candidates.   Completions  candidates are
+atoms that
+
+  - Are built-in or referenced from some static datastructure
+  - All characters are legal characters for unquoted atoms
+  - The atom is at most 80 characters long
+*/
+
+static
+PRED_IMPL("$atom_completions", 2, atom_completions, 0)
+{ PRED_LD
+  term_t prefix = A1;
+  term_t alternatives = A2;
+
+  PL_chars_t p_text;
   struct match altv[ALT_MAX];
   int altn;
   int i;
   term_t alts = PL_copy_term_ref(alternatives);
   term_t head = PL_new_term_ref();
 
-  if ( !PL_get_chars(prefix, &p, CVT_ALL|CVT_EXCEPTION) )
-    fail;
-  strcpy(buf, p);
+  if ( !PL_get_text(prefix, &p_text, CVT_ALL|CVT_EXCEPTION) )
+    return FALSE;
 
-  extend_alternatives(buf, altv, &altn);
+  if ( !extend_alternatives(&p_text, altv, &altn) )
+    return FALSE;			/* interrupt */
 
   for(i=0; i<altn; i++)
   { if ( !PL_unify_list(alts, head, alts) ||
 	 !PL_unify_atom(head, altv[i].name->atom) )
-      fail;
+      return FALSE;
   }
 
   return PL_unify_nil(alts);
@@ -1455,38 +1559,6 @@ thread.
 static pthread_key_t key;
 #endif
 
-#define is_signalled() unlikely(LD && LD->signal.pending != 0)
-
-static int
-alnum_text(PL_chars_t *txt)
-{ switch(txt->encoding)
-  { case ENC_ISO_LATIN_1:
-    { const unsigned char *s = (const unsigned char *)txt->text.t;
-      const unsigned char *e = &s[txt->length];
-
-      for(; s<e; s++)
-      { if ( !isAlpha(*s) )
-	  return FALSE;
-      }
-      return TRUE;
-    }
-    case ENC_WCHAR:
-    { const pl_wchar_t *s = (const pl_wchar_t*)txt->text.w;
-      const pl_wchar_t *e = &s[txt->length];
-
-      for(; s<e; s++)
-      { if ( !isAlphaW(*s) )
-	  return FALSE;
-      }
-      return TRUE;
-    }
-    default:
-      assert(0);
-      return FALSE;
-  }
-}
-
-
 static int
 atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
 { GET_LD
@@ -1524,10 +1596,11 @@ atom_generator(PL_chars_t *prefix, PL_chars_t *hit, int state)
       if ( is_signalled() )		/* Notably allow windows version */
 	PL_handle_signals();		/* to break out on ^C */
 
-      if ( a && get_atom_ptr_text(a, hit) &&
-	 hit->length < ALT_SIZ &&
-	 PL_cmp_text(prefix, 0, hit, 0, prefix->length) == 0 &&
-	 alnum_text(hit) )
+      if ( a && completion_candidate(a) &&
+	   get_atom_ptr_text(a, hit) &&
+	   hit->length < ALT_SIZ &&
+	   PL_cmp_text(prefix, 0, hit, 0, prefix->length) == 0 &&
+	   is_identifier_text(hit) )
       {
 #ifdef O_PLMT
         pthread_setspecific(key, (void *)(index+1));
@@ -1605,4 +1678,6 @@ BeginPredDefs(atom)
   PRED_DEF("current_atom", 1, current_atom, PL_FA_NONDETERMINISTIC)
   PRED_DEF("blob", 2, blob, 0)
   PRED_DEF("$atom_references", 2, atom_references, 0)
+  PRED_DEF("$atom_completions", 2, atom_completions, 0)
+  PRED_DEF("$complete_atom", 3, complete_atom, 0)
 EndPredDefs
