@@ -368,11 +368,14 @@ shareDefinition(Definition def)
 }
 
 
-void
+int
 unshareDefinition(Definition def)
-{ LOCKDEF(def);
-  def->shared--;
+{ int times;
+  LOCKDEF(def);
+  times = --def->shared;
   UNLOCKDEF(def);
+
+  return times;
 }
 
 
@@ -410,10 +413,15 @@ get_arity(term_t t, int extra, int maxarity, int *arity)
 int
 get_functor(term_t descr, functor_t *fdef, Module *m, term_t h, int how)
 { GET_LD
-  term_t head = PL_new_term_ref();
+  term_t head;
   int dcgpi=FALSE;
 
-  PL_strip_module(descr, m, head);
+  if ( !(how&GP_NOT_QUALIFIED) )
+  { head = PL_new_term_ref();
+    PL_strip_module(descr, m, head);
+  } else
+  { head = descr;
+  }
 
   if ( PL_is_functor(head, FUNCTOR_divide2) ||
        (dcgpi=PL_is_functor(head, FUNCTOR_gdiv2)) )
@@ -899,7 +907,7 @@ assertProcedure(Procedure proc, Clause clause, int where ARG_LD)
   word key;
   ClauseRef cref;
 
-  argKey(clause->codes, 0, FALSE, &key);
+  argKey(clause->codes, 0, &key);
   cref = newClauseRef(clause, key);
 
   if ( def->references && (debugstatus.styleCheck & DYNAMIC_STYLE) )
@@ -964,9 +972,10 @@ abolishProcedure(Procedure proc, Module module)
   { Definition ndef	     = allocHeapOrHalt(sizeof(struct definition));
 
     memset(ndef, 0, sizeof(*ndef));
-    proc->definition         = ndef;
     ndef->functor            = def->functor; /* should be merged with */
     ndef->module             = module;	     /* lookupProcedure()!! */
+    ndef->codes		     = SUPERVISOR(virgin);
+    proc->definition         = ndef;
     resetProcedure(proc, TRUE);
   } else if ( true(def, P_FOREIGN) )	/* foreign: make normal */
   { def->impl.clauses.first_clause = def->impl.clauses.last_clause = NULL;
@@ -1623,6 +1632,7 @@ pl_garbage_collect_clauses(void)
     LOCK();
     PL_LOCK(L_GC);
     PL_LOCK(L_THREAD);
+    PL_LOCK(L_STOPTHEWORLD);
     blockSignals(&set);
 
 					/* sanity-check */
@@ -1673,6 +1683,7 @@ pl_garbage_collect_clauses(void)
 #endif
 
     unblockSignals(&set);
+    PL_UNLOCK(L_STOPTHEWORLD);
     PL_UNLOCK(L_THREAD);
     PL_UNLOCK(L_GC);
     UNLOCK();
@@ -1789,7 +1800,9 @@ autoImport(functor_t f, Module m)
     return proc->definition;
 
   for(c=m->supers; c; c=c->next)
-  { if ( (def = autoImport(f, c->value)) )
+  { Module s = c->value;
+
+    if ( (def = autoImport(f, s)) )
       goto found;
   }
   return NULL;
@@ -1797,26 +1810,35 @@ autoImport(functor_t f, Module m)
 found:
   if ( proc == NULL )			/* Create header if not there */
     proc = lookupProcedure(f, m);
-					/* Safe? See above */
-					/* TBD: find something better! */
-  odef = proc->definition;
-  proc->definition = def;
-  shareDefinition(def);
-  unshareDefinition(odef);
 
+					/* Now, take the lock also used */
+					/* by lookupProcedure().  Note */
+					/* that another thread may have */
+					/* done the job for us. */
+  LOCKMODULE(m);
+  if ( (odef=proc->definition) != def )	/* Nope, we must link the def */
+  { proc->definition = def;
+    shareDefinition(def);
+    unshareDefinition(odef);
+
+    if ( unshareDefinition(odef) == 0 )
+    {
 #ifdef O_PLMT
-  PL_LOCK(L_THREAD);
-  if ( (GD->statistics.threads_created -
-	GD->statistics.threads_finished) == 1 )
-  { assert(false(proc->definition, P_DIRTYREG));
-    freeHeap(odef, sizeof(struct definition));
-  } else
-  { GC_LINGER(odef);
-  }
-  PL_UNLOCK(L_THREAD);
+      PL_LOCK(L_THREAD);
+      if ( (GD->statistics.threads_created -
+	    GD->statistics.threads_finished) == 1 )
+      { assert(false(proc->definition, P_DIRTYREG));
+	freeHeap(odef, sizeof(struct definition));
+      } else
+      { GC_LINGER(odef);
+      }
+      PL_UNLOCK(L_THREAD);
 #else
-  freeHeap(odef, sizeof(struct definition));
+      freeHeap(odef, sizeof(struct definition));
 #endif
+    }
+  }
+  UNLOCKMODULE(m);
 
   return def;
 }
@@ -2265,26 +2287,28 @@ typedef struct patt_mask
 #define TRACE_ANY (TRACE_CALL|TRACE_REDO|TRACE_EXIT|TRACE_FAIL)
 
 static const patt_mask patt_masks[] =
-{ { ATOM_dynamic,	P_DYNAMIC },
-  { ATOM_multifile,	P_MULTIFILE },
-  { ATOM_locked,	P_LOCKED },
-  { ATOM_system,	P_LOCKED },		/* compatibility */
-  { ATOM_spy,		SPY_ME },
-  { ATOM_trace,		TRACE_ME },
-  { ATOM_trace_call,	TRACE_CALL },
-  { ATOM_trace_redo,	TRACE_REDO },
-  { ATOM_trace_exit,	TRACE_EXIT },
-  { ATOM_trace_fail,	TRACE_FAIL },
-  { ATOM_trace_any,	TRACE_ANY },
-  { ATOM_hide_childs,	HIDE_CHILDS },
-  { ATOM_transparent,	P_TRANSPARENT },
-  { ATOM_discontiguous,	P_DISCONTIGUOUS },
-  { ATOM_volatile,	P_VOLATILE },
-  { ATOM_thread_local,	P_THREAD_LOCAL },
-  { ATOM_noprofile,	P_NOPROFILE },
-  { ATOM_iso,		P_ISO },
-  { ATOM_public,	P_PUBLIC },
-  { (atom_t)0,		0 }
+{ { ATOM_dynamic,	   P_DYNAMIC },
+  { ATOM_multifile,	   P_MULTIFILE },
+  { ATOM_locked,	   P_LOCKED },
+  { ATOM_system,	   P_LOCKED },		/* compatibility */
+  { ATOM_spy,		   SPY_ME },
+  { ATOM_trace,		   TRACE_ME },
+  { ATOM_trace_call,	   TRACE_CALL },
+  { ATOM_trace_redo,	   TRACE_REDO },
+  { ATOM_trace_exit,	   TRACE_EXIT },
+  { ATOM_trace_fail,	   TRACE_FAIL },
+  { ATOM_trace_any,	   TRACE_ANY },
+  { ATOM_hide_childs,	   HIDE_CHILDS },
+  { ATOM_transparent,	   P_TRANSPARENT },
+  { ATOM_discontiguous,	   P_DISCONTIGUOUS },
+  { ATOM_volatile,	   P_VOLATILE },
+  { ATOM_thread_local,	   P_THREAD_LOCAL },
+  { ATOM_noprofile,	   P_NOPROFILE },
+  { ATOM_iso,		   P_ISO },
+  { ATOM_public,	   P_PUBLIC },
+  { ATOM_non_terminal,	   P_NON_TERMINAL },
+  { ATOM_quasi_quotation_syntax, P_QUASI_QUOTATION_SYNTAX },
+  { (atom_t)0,		   0 }
 };
 
 static unsigned int
@@ -3101,6 +3125,7 @@ unloadFile(SourceFile sf)
 
   LOCK();
   PL_LOCK(L_THREAD);
+  PL_LOCK(L_STOPTHEWORLD);
   blockSignals(&set);
 
   GD->procedures.active_marked = 0;
@@ -3174,6 +3199,7 @@ unloadFile(SourceFile sf)
   delAllModulesSourceFile__unlocked(sf);
 
   unblockSignals(&set);
+  PL_UNLOCK(L_STOPTHEWORLD);
   PL_UNLOCK(L_THREAD);
   UNLOCK();
 
@@ -3476,7 +3502,7 @@ listGenerations(Definition def)
     { unsigned int i;
 
       Sdprintf("\nHash %sindex for arg %d (%d dirty)\n",
-	       ci->is_list ? "list-" : "", ci->arg, ci->dirty);
+	       ci->is_list ? "list-" : "", ci->args[0], ci->dirty);
 
       for(i=0; i<ci->buckets; i++)
       { if ( !ci->entries[i].head &&
