@@ -171,6 +171,7 @@ bits) as well as machines with different byte order.
 #define XR_BLOB	   9			/* a typed atom (blob) */
 #define XR_BLOB_TYPE 10			/* name of atom-type declaration */
 #define XR_STRING_UTF8 11		/* Wide character string */
+#define XR_NULL	  12			/* NULL pointer */
 
 #define PRED_SYSTEM	 0x01		/* system predicate */
 #define PRED_HIDE_CHILDS 0x02		/* hide my childs */
@@ -213,10 +214,11 @@ typedef struct wic_state
   Table	savedXRTable;			/* saved XR entries */
   intptr_t savedXRTableId;		/* next id to hand out */
 
-  int	     has_source_marks;
   SourceMark source_mark_head;		/* Locations of sources */
   SourceMark source_mark_tail;
+  int	     has_source_marks;
 
+  int	     load_nesting;		/* Nesting level of loadPart() */
   qlf_state *load_state;		/* current load-state */
 
   xr_table *XR;				/* external references */
@@ -732,6 +734,8 @@ loadXRc(wic_state *state, int c ARG_LD)
 
       break;
     }
+    case XR_NULL:
+      return 0;
     default:
     { xr = 0;				/* make gcc happy */
       fatalError("Illegal XR entry at index %d: %c", Stell(fd)-1, c);
@@ -1004,8 +1008,14 @@ loadStatement(wic_state *state, int c, int skip ARG_LD)
     }
 
     case 'Q':
-      return loadPart(state, NULL, skip PASS_LD);
+    { bool rc;
 
+      state->load_nesting++;
+      rc = loadPart(state, NULL, skip PASS_LD);
+      state->load_nesting--;
+
+      return rc;
+    }
     case 'M':
       return loadInModule(state, skip PASS_LD);
 
@@ -1149,6 +1159,7 @@ loadPredicate(wic_state *state, int skip ARG_LD)
 	      case CA1_INTEGER:
 	      case CA1_JUMP:
 	      case CA1_VAR:
+	      case CA1_FVAR:
 	      case CA1_CHP:
 		*bp++ = (intptr_t)getInt64(fd);
 		break;
@@ -1457,8 +1468,10 @@ loadPart(wic_state *state, Module *module, int skip ARG_LD)
 
     switch(c)
     { case 'X':
-      { if ( !GD->bootsession )
+      { if ( !GD->bootsession && state->load_nesting > 0 )
+	{ /* top-level '$run_initialization'/1 is called from boot/init.pl */
 	  runInitialization(state->currentSource);
+	}
 	LD->modules.source = om;
 	state->currentSource  = of;
 	debugstatus.styleCheck = stchk;
@@ -1833,11 +1846,15 @@ saveXRModule(wic_state *state, Module m ARG_LD)
   if ( savedXRPointer(state, m) )
     return;
 
-  Sputc(XR_MODULE, fd);
-  DEBUG(MSG_QLF_XR,
-	Sdprintf("XR(%d) = module %s\n",
-		 state->savedXRTableId, stringAtom(m->name)));
-  saveXR(state, m->name);
+  if ( m )
+  { Sputc(XR_MODULE, fd);
+    DEBUG(MSG_QLF_XR,
+	  Sdprintf("XR(%d) = module %s\n",
+		   state->savedXRTableId, stringAtom(m->name)));
+    saveXR(state, m->name);
+  } else
+  { Sputc(XR_NULL, fd);
+  }
 }
 
 
@@ -2027,6 +2044,7 @@ saveWicClause(wic_state *state, Clause clause)
 	case CA1_INTEGER:
 	case CA1_JUMP:
 	case CA1_VAR:
+	case CA1_FVAR:
 	case CA1_CHP:
 	{ putNum(*bp++, fd);
 	  break;
